from slither.core import variables
from slither.core.declarations import (
    SolidityVariable,
    SolidityVariableComposed,
    Structure,
    Enum,
    Contract,
)
from slither.core import solidity_types
from slither.slithir import operations
from slither.slithir import variables as SlitherIRVariable


def ntype(_type: solidity_types.Type | str) -> str:
    if isinstance(_type, solidity_types.ElementaryType):
        _type = str(_type)
    elif isinstance(_type, solidity_types.ArrayType):
        if isinstance(_type.type, solidity_types.ElementaryType):
            _type = str(_type)
        else:
            _type = "user_defined_array"
    elif isinstance(_type, Structure):
        _type = str(_type)
    elif isinstance(_type, Enum):
        _type = str(_type)
    elif isinstance(_type, solidity_types.MappingType):
        _type = str(_type)
    elif isinstance(_type, solidity_types.UserDefinedType):
        if isinstance(_type.type, Contract):
            _type = f"contract({_type.type.name})"
        elif isinstance(_type.type, Structure):
            _type = f"struct({_type.type.name})"
        elif isinstance(_type.type, Enum):
            _type = f"enum({_type.type.name})"
    else:
        _type = str(_type)

    _type = _type.replace(" memory", "")
    _type = _type.replace(" storage ref", "")

    if "struct" in _type:
        return "struct"
    if "enum" in _type:
        return "enum"
    if "tuple" in _type:
        return "tuple"
    if "contract" in _type:
        return "contract"
    if "mapping" in _type:
        return "mapping"
    return _type.replace(" ", "_")


def encode_var_for_compare(var: variables.Variable | SolidityVariable) -> str:
    # variables
    if isinstance(var, SlitherIRVariable.Constant):
        return f"constant({ntype(var.type)},{var.value})"
    if isinstance(var, SolidityVariableComposed):
        return f"solidity_variable_composed({var.name})"
    if isinstance(var, SolidityVariable):
        return f"solidity_variable{var.name}"
    if isinstance(var, SlitherIRVariable.TemporaryVariable):
        return "temporary_variable"
    if isinstance(var, SlitherIRVariable.ReferenceVariable):
        return f"reference({ntype(var.type)})"
    if isinstance(var, variables.LocalVariable):
        return f"local_solc_variable({ntype(var.type)},{var.location})"
    if isinstance(var, variables.StateVariable):
        if var.is_stored:
            try:
                slot, _ = var.contract.compilation_unit.storage_layout_of(var.contract, var)
            except KeyError:
                slot = var.name
        else:
            slot = var.name
        return f"state_solc_variable({ntype(var.type)},{slot})"
    if isinstance(var, variables.LocalVariableInitFromTuple):
        return "local_variable_init_tuple"
    if isinstance(var, SlitherIRVariable.TupleVariable):
        return "tuple_variable"

    # default
    return ""


def encode_ir_for_upgradeability_compare(ir: operations.Operation) -> str:
    # operations
    if isinstance(ir, operations.Assignment):
        return f"({encode_var_for_compare(ir.lvalue)}):=({encode_var_for_compare(ir.rvalue)})"
    if isinstance(ir, operations.Index):
        return f"index({ntype(ir.variable_right.type)})"
    if isinstance(ir, operations.Member):
        return "member"  # .format(ntype(ir._type))
    if isinstance(ir, operations.Length):
        return "length"
    if isinstance(ir, operations.Binary):
        return f"binary({encode_var_for_compare(ir.variable_left)}{ir.type}{encode_var_for_compare(ir.variable_right)})"
    if isinstance(ir, operations.Unary):
        return f"unary({ir.type!s})"
    if isinstance(ir, operations.Condition):
        return f"condition({encode_var_for_compare(ir.value)})"
    if isinstance(ir, operations.NewStructure):
        return "new_structure"
    if isinstance(ir, operations.NewContract):
        return "new_contract"
    if isinstance(ir, operations.NewArray):
        return f"new_array({ntype(ir.array_type)})"
    if isinstance(ir, operations.NewElementaryType):
        return f"new_elementary({ntype(ir.type)})"
    if isinstance(ir, operations.Delete):
        return f"delete({encode_var_for_compare(ir.lvalue)},{encode_var_for_compare(ir.variable)})"
    if isinstance(ir, operations.SolidityCall):
        return f"solidity_call({ir.function.full_name})"
    if isinstance(ir, operations.InternalCall):
        return f"internal_call({ntype(ir.type_call)})"
    if isinstance(ir, operations.EventCall):  # is this useful?
        return "event"
    if isinstance(ir, operations.LibraryCall):
        return "library_call"
    if isinstance(ir, operations.InternalDynamicCall):
        return "internal_dynamic_call"
    if isinstance(ir, operations.HighLevelCall):  # TODO: improve
        return "high_level_call"
    if isinstance(ir, operations.LowLevelCall):  # TODO: improve
        return "low_level_call"
    if isinstance(ir, operations.TypeConversion):
        return f"type_conversion({ntype(ir.type)})"
    if isinstance(ir, operations.Return):  # this can be improved using values
        return "return"  # .format(ntype(ir.type))
    if isinstance(ir, operations.Transfer):
        return f"transfer({encode_var_for_compare(ir.call_value)})"
    if isinstance(ir, operations.Send):
        return f"send({encode_var_for_compare(ir.call_value)})"
    if isinstance(ir, operations.Unpack):  # TODO: improve
        return "unpack"
    if isinstance(ir, operations.InitArray):  # TODO: improve
        return "init_array"

    # default
    return ""


def encode_ir_for_halstead(ir: operations.Operation) -> str:
    # operations
    if isinstance(ir, operations.Assignment):
        return "assignment"
    if isinstance(ir, operations.Index):
        return "index"
    if isinstance(ir, operations.Member):
        return "member"  # .format(ntype(ir._type))
    if isinstance(ir, operations.Length):
        return "length"
    if isinstance(ir, operations.Binary):
        return f"binary({ir.type!s})"
    if isinstance(ir, operations.Unary):
        return f"unary({ir.type!s})"
    if isinstance(ir, operations.Condition):
        return f"condition({encode_var_for_compare(ir.value)})"
    if isinstance(ir, operations.NewStructure):
        return "new_structure"
    if isinstance(ir, operations.NewContract):
        return "new_contract"
    if isinstance(ir, operations.NewArray):
        return f"new_array({ntype(ir.array_type)})"
    if isinstance(ir, operations.NewElementaryType):
        return f"new_elementary({ntype(ir.type)})"
    if isinstance(ir, operations.Delete):
        return "delete"
    if isinstance(ir, operations.SolidityCall):
        return f"solidity_call({ir.function.full_name})"
    if isinstance(ir, operations.InternalCall):
        return f"internal_call({ntype(ir.type_call)})"
    if isinstance(ir, operations.EventCall):  # is this useful?
        return "event"
    if isinstance(ir, operations.LibraryCall):
        return "library_call"
    if isinstance(ir, operations.InternalDynamicCall):
        return "internal_dynamic_call"
    if isinstance(ir, operations.HighLevelCall):  # TODO: improve
        return "high_level_call"
    if isinstance(ir, operations.LowLevelCall):  # TODO: improve
        return "low_level_call"
    if isinstance(ir, operations.TypeConversion):
        return f"type_conversion({ntype(ir.type)})"
    if isinstance(ir, operations.Return):  # this can be improved using values
        return "return"  # .format(ntype(ir.type))
    if isinstance(ir, operations.Transfer):
        return "transfer"
    if isinstance(ir, operations.Send):
        return "send"
    if isinstance(ir, operations.Unpack):  # TODO: improve
        return "unpack"
    if isinstance(ir, operations.InitArray):  # TODO: improve
        return "init_array"
    # default
    raise NotImplementedError(f"encode_ir_for_halstead: {ir}")



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import (
    Binary,
    BinaryType,
)

from slither.core.declarations import Function
from slither.utils.output import Output


class TautologicalCompare(AbstractDetector):
    """
    Same variable comparison detector
    """

    ARGUMENT = "tautological-compare"
    HELP = "Comparing a variable to itself always returns true or false, depending on comparison"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#tautological-compare"

    WIKI_TITLE = "Tautological compare"
    WIKI_DESCRIPTION = "A variable compared to itself is probably an error as it will always return `true` for `==`, `>=`, `<=` and always `false` for `<`, `>` and `!=`."
    WIKI_EXPLOIT_SCENARIO = """
```solidity
    function check(uint a) external returns(bool){
        return (a >= a);
    }
```
`check` always return true."""

    WIKI_RECOMMENDATION = "Remove comparison or compare to different value."

    def _check_function(self, f: Function) -> list[Output]:
        affected_nodes = set()
        for node in f.nodes:
            for ir in node.irs:
                if isinstance(ir, Binary):
                    if ir.type in [
                        BinaryType.GREATER,
                        BinaryType.GREATER_EQUAL,
                        BinaryType.LESS,
                        BinaryType.LESS_EQUAL,
                        BinaryType.EQUAL,
                        BinaryType.NOT_EQUAL,
                    ]:
                        if ir.variable_left == ir.variable_right:
                            affected_nodes.add(node)

        results = []
        for n in affected_nodes:
            info: DETECTOR_INFO = [f, " compares a variable to itself:\n\t", n, "\n"]
            res = self.generate_result(info)
            results.append(res)
        return results

    def _detect(self):
        results = []

        for f in self.compilation_unit.functions_and_modifiers:
            results.extend(self._check_function(f))

        return results



from slither.core.declarations import SolidityFunction, Function
from slither.core.declarations.contract import Contract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


def _can_be_destroyed(contract: Contract) -> list[Function]:
    targets = []
    for f in contract.functions_entry_points:
        found = False
        for ir in f.all_low_level_calls():
            if ir.function_name in ["delegatecall", "codecall"]:
                targets.append(f)
                found = True
                break

        if not found:
            for ir in f.all_solidity_calls():
                if ir.function in [
                    SolidityFunction("suicide(address)"),
                    SolidityFunction("selfdestruct(address)"),
                ]:
                    targets.append(f)
                    break

    return targets


def _has_initializing_protection(functions: list[Function]) -> bool:
    # Detects "initializer" constructor modifiers and "_disableInitializers()" constructor internal calls
    # https://docs.openzeppelin.com/upgrades-plugins/1.x/writing-upgradeable#initializing_the_implementation_contract

    for f in functions:
        for m in f.modifiers:
            if m.name == "initializer":
                return True
        for ir in f.all_internal_calls():
            if ir.function.name == "_disableInitializers":
                return True

    # to avoid future FPs in different modifier + function naming implementations, we can also implement a broader check for state var "_initialized" being written to in the constructor
    #   though this is still subject to naming false positives...
    return False


def _whitelisted_modifiers(f: Function) -> bool:
    # The onlyProxy modifier prevents calling the implementation contract (must be delegatecall)
    #  https://github.com/OpenZeppelin/openzeppelin-contracts-upgradeable/blob/3dec82093ea4a490d63aab3e925fed4f692909e8/contracts/proxy/utils/UUPSUpgradeable.sol#L38-L42
    return "onlyProxy" not in [modifier.name for modifier in f.modifiers]


def _initialize_functions(contract: Contract) -> list[Function]:
    return list(
        filter(
            _whitelisted_modifiers,
            [
                f
                for f in contract.functions
                if any((m.name in ["initializer", "reinitializer"]) for m in f.modifiers)
            ],
        )
    )


class UnprotectedUpgradeable(AbstractDetector):
    ARGUMENT = "unprotected-upgrade"
    HELP = "Unprotected upgradeable contract"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unprotected-upgradeable-contract"

    WIKI_TITLE = "Unprotected upgradeable contract"
    WIKI_DESCRIPTION = """Detects logic contract that can be destructed."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Buggy is Initializable{
    address payable owner;

    function initialize() external initializer{
        require(owner == address(0));
        owner = msg.sender;
    }
    function kill() external{
        require(msg.sender == owner);
        selfdestruct(owner);
    }
}
```
Buggy is an upgradeable contract. Anyone can call initialize on the logic contract, and destruct the contract.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        """Add a constructor to ensure `initialize` cannot be called on the logic contract."""
    )

    def _detect(self) -> list[Output]:
        results = []

        for contract in self.compilation_unit.contracts_derived:
            if contract.is_upgradeable:
                if not _has_initializing_protection(contract.constructors):
                    functions_that_can_destroy = _can_be_destroyed(contract)
                    if functions_that_can_destroy:
                        initialize_functions = _initialize_functions(contract)

                        vars_init_ = [
                            init.all_state_variables_written() for init in initialize_functions
                        ]
                        vars_init = [item for sublist in vars_init_ for item in sublist]

                        vars_init_in_constructors_ = [
                            f.all_state_variables_written() for f in contract.constructors
                        ]
                        vars_init_in_constructors = [
                            item for sublist in vars_init_in_constructors_ for item in sublist
                        ]
                        if vars_init and (set(vars_init) - set(vars_init_in_constructors)):
                            info: DETECTOR_INFO = [
                                contract,
                                " is an upgradeable contract that does not protect its initialize functions: ",
                            ]
                            info += initialize_functions
                            info += [
                                ". Anyone can delete the contract with: ",
                            ]
                            info += functions_that_can_destroy

                            res = self.generate_result(info)
                            results.append(res)

        return results



"""
This class is used for the SSA version of slithIR
It is similar to the non-SSA version of slithIR
as the ReferenceVariable are in SSA form in both version
"""

from slither.slithir.variables.reference import ReferenceVariable
from slither.slithir.variables.tuple import TupleVariable


class ReferenceVariableSSA(ReferenceVariable):
    def __init__(self, reference: ReferenceVariable) -> None:
        super().__init__(reference.node, reference.index)

        self._non_ssa_version = reference

    @property
    def non_ssa_version(self) -> ReferenceVariable | TupleVariable:
        return self._non_ssa_version



"""
Module detecting send to arbitrary address

To avoid FP, it does not report:
    - If msg.sender is used as index (withdraw situation)
    - If the function is protected
    - If the value sent is msg.value (repay situation)
    - If there is a call to transferFrom

TODO: dont report if the value is tainted by msg.value
"""

from typing import Any

from slither.analyses.data_dependency.data_dependency import is_tainted, is_dependent
from slither.core.cfg.node import Node
from slither.core.declarations import Function, Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.solidity_variables import (
    SolidityFunction,
    SolidityVariableComposed,
    SolidityVariable,
)
from slither.core.variables import Variable
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import (
    HighLevelCall,
    Index,
    LowLevelCall,
    Send,
    SolidityCall,
    Transfer,
)
from slither.core.variables.state_variable import StateVariable


from slither.utils.output import Output


def arbitrary_send(func: Function) -> bool | list[Node]:
    if func.is_protected():
        return []

    ret: list[Node] = []
    for node in func.nodes:
        func = node.function
        deps_target: Contract | Function = (
            func.contract if isinstance(func, FunctionContract) else func
        )
        for ir in node.irs:
            if isinstance(ir, SolidityCall):
                if ir.function == SolidityFunction("ecrecover(bytes32,uint8,bytes32,bytes32)"):
                    return False
            if isinstance(ir, Index):
                if ir.variable_right == SolidityVariableComposed("msg.sender"):
                    return False
                if is_dependent(
                    ir.variable_right,
                    SolidityVariableComposed("msg.sender"),
                    deps_target,
                ):
                    return False
            if isinstance(ir, (HighLevelCall, LowLevelCall, Transfer, Send)):
                if isinstance(ir, (HighLevelCall)):
                    if isinstance(ir.function, Function):
                        if ir.function.full_name == "transferFrom(address,address,uint256)":
                            return False
                if ir.call_value is None:
                    continue
                if ir.call_value == SolidityVariableComposed("msg.value"):
                    continue
                if isinstance(ir.destination, StateVariable) and ir.destination.is_immutable:
                    continue
                if is_dependent(
                    ir.call_value,
                    SolidityVariableComposed("msg.value"),
                    node,
                ):
                    continue

                if isinstance(ir.destination, (Variable, SolidityVariable)):
                    if is_tainted(ir.destination, node):
                        ret.append(node)

    return ret


def detect_arbitrary_send(
    contract: Contract,
) -> list[tuple[FunctionContract, list[Node]] | Any]:
    """
        Detect arbitrary send
    Args:
        contract (Contract)
    Returns:
        list((Function), (list (Node)))
    """
    ret = []
    for f in [f for f in contract.functions if f.contract_declarer == contract]:
        nodes = arbitrary_send(f)
        if nodes:
            ret.append((f, nodes))
    return ret


class ArbitrarySendEth(AbstractDetector):
    ARGUMENT = "arbitrary-send-eth"
    HELP = "Functions that send Ether to arbitrary destinations"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#functions-that-send-ether-to-arbitrary-destinations"

    WIKI_TITLE = "Functions that send Ether to arbitrary destinations"
    WIKI_DESCRIPTION = "Unprotected call to a function sending Ether to an arbitrary address."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract ArbitrarySendEth{
    address destination;
    function setDestination(){
        destination = msg.sender;
    }

    function withdraw() public{
        destination.transfer(this.balance);
    }
}
```
Bob calls `setDestination` and `withdraw`. As a result he withdraws the contract's balance."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Ensure that an arbitrary user cannot withdraw unauthorized funds."

    def _detect(self) -> list[Output]:
        """"""
        results = []

        for c in self.contracts:
            arbitrary_send_result = detect_arbitrary_send(c)
            for func, nodes in arbitrary_send_result:
                info = [func, " sends eth to arbitrary user\n"]
                info += ["\tDangerous calls:\n"]

                # sort the nodes to get deterministic results
                nodes.sort(key=lambda x: x.node_id)

                for node in nodes:
                    info += ["\t- ", node, "\n"]

                res = self.generate_result(info)

                results.append(res)

        return results



"""
Halstead complexity metrics
https://en.wikipedia.org/wiki/Halstead_complexity_measures

12 metrics based on the number of unique operators and operands:

Core metrics:
n1 = the number of distinct operators
n2 = the number of distinct operands
N1 = the total number of operators
N2 = the total number of operands

Extended metrics1:
n = n1 + n2  # Program vocabulary
N = N1 + N2  # Program length
S = n1 * log2(n1) + n2 * log2(n2) # Estimated program length
V = N * log2(n) # Volume

Extended metrics2:
D = (n1 / 2) * (N2 / n2) # Difficulty
E = D * V # Effort
T = E / 18 seconds # Time required to program
B = (E^(2/3)) / 3000 # Number of delivered bugs


"""

import math
from collections import OrderedDict
from dataclasses import dataclass, field

from slither.core.declarations import Contract
from slither.slithir.variables.temporary import TemporaryVariable
from slither.utils.encoding import encode_ir_for_halstead
from slither.utils.myprettytable import make_pretty_table, MyPrettyTable


@dataclass(slots=True)
class HalsteadContractMetrics:
    """Class to hold the Halstead metrics for a single contract."""

    contract: Contract
    all_operators: list[str] = field(default_factory=list)
    all_operands: list[str] = field(default_factory=list)
    n1: int = 0
    n2: int = 0
    N1: int = 0
    N2: int = 0
    n: int = 0
    N: int = 0
    S: float = 0
    V: float = 0
    D: float = 0
    E: float = 0
    T: float = 0
    B: float = 0

    def __post_init__(self) -> None:
        """Operators and operands can be passed in as constructor args to avoid computing
        them based on the contract. Useful for computing metrics for ALL_CONTRACTS"""

        if len(self.all_operators) == 0:
            if not hasattr(self.contract, "functions"):
                return
            self.populate_operators_and_operands()
        if len(self.all_operators) > 0:
            self.compute_metrics()

    def to_dict(self) -> dict[str, float]:
        """Return the metrics as a dictionary."""
        return OrderedDict(
            {
                "Total Operators": self.N1,
                "Unique Operators": self.n1,
                "Total Operands": self.N2,
                "Unique Operands": self.n2,
                "Vocabulary": str(self.n1 + self.n2),
                "Program Length": str(self.N1 + self.N2),
                "Estimated Length": f"{self.S:.0f}",
                "Volume": f"{self.V:.0f}",
                "Difficulty": f"{self.D:.0f}",
                "Effort": f"{self.E:.0f}",
                "Time": f"{self.T:.0f}",
                "Estimated Bugs": f"{self.B:.3f}",
            }
        )

    def populate_operators_and_operands(self) -> None:
        """Populate the operators and operands lists."""
        operators = []
        operands = []

        for func in self.contract.functions:
            for node in func.nodes:
                for operation in node.irs:
                    # use operation.expression.type to get the unique operator type
                    encoded_operator = encode_ir_for_halstead(operation)
                    operators.append(encoded_operator)

                    # use operation.used to get the operands of the operation ignoring the temporary variables
                    operands.extend(
                        [op for op in operation.used if not isinstance(op, TemporaryVariable)]
                    )
        self.all_operators.extend(operators)
        self.all_operands.extend(operands)

    def compute_metrics(self, all_operators=None, all_operands=None) -> None:
        """Compute the Halstead metrics."""
        if all_operators is None:
            all_operators = self.all_operators
            all_operands = self.all_operands

        # core metrics
        self.n1 = len(set(all_operators))
        self.n2 = len(set(all_operands))
        self.N1 = len(all_operators)
        self.N2 = len(all_operands)
        if any(number <= 0 for number in [self.n1, self.n2, self.N1, self.N2]):
            raise ValueError("n1 and n2 must be greater than 0")

        # extended metrics 1
        self.n = self.n1 + self.n2
        self.N = self.N1 + self.N2
        self.S = self.n1 * math.log2(self.n1) + self.n2 * math.log2(self.n2)
        self.V = self.N * math.log2(self.n)

        # extended metrics 2
        self.D = (self.n1 / 2) * (self.N2 / self.n2)
        self.E = self.D * self.V
        self.T = self.E / 18
        self.B = (self.E ** (2 / 3)) / 3000


@dataclass(slots=True)
class SectionInfo:
    """Class to hold the information for a section of the report."""

    title: str
    pretty_table: MyPrettyTable
    txt: str


@dataclass(slots=True)
class HalsteadMetrics:
    """Class to hold the Halstead metrics for all contracts. Contains methods useful for reporting.

    There are 3 sections in the report:
    1. Core metrics (n1, n2, N1, N2)
    2. Extended metrics 1 (n, N, S, V)
    3. Extended metrics 2 (D, E, T, B)

    """

    contracts: list[Contract] = field(default_factory=list)
    contract_metrics: OrderedDict = field(default_factory=OrderedDict)
    title: str = "Halstead complexity metrics"
    full_text: str = ""
    core: SectionInfo = field(default=SectionInfo)
    extended1: SectionInfo = field(default=SectionInfo)
    extended2: SectionInfo = field(default=SectionInfo)
    CORE_KEYS = (
        "Total Operators",
        "Unique Operators",
        "Total Operands",
        "Unique Operands",
    )
    EXTENDED1_KEYS = (
        "Vocabulary",
        "Program Length",
        "Estimated Length",
        "Volume",
    )
    EXTENDED2_KEYS = (
        "Difficulty",
        "Effort",
        "Time",
        "Estimated Bugs",
    )
    SECTIONS: tuple[tuple[str, str, tuple[str]]] = (
        ("Core", "core", CORE_KEYS),
        ("Extended 1/2", "extended1", EXTENDED1_KEYS),
        ("Extended 2/2", "extended2", EXTENDED2_KEYS),
    )

    def __post_init__(self) -> None:
        # Compute the metrics for each contract and for all contracts.
        self.update_contract_metrics()
        self.add_all_contracts_metrics()
        self.update_reporting_sections()

    def update_contract_metrics(self) -> None:
        for contract in self.contracts:
            self.contract_metrics[contract.name] = HalsteadContractMetrics(contract=contract)

    def add_all_contracts_metrics(self) -> None:
        # If there are more than 1 contract, compute the metrics for all contracts.
        if len(self.contracts) <= 1:
            return
        all_operators = [
            operator
            for contract in self.contracts
            for operator in self.contract_metrics[contract.name].all_operators
        ]
        all_operands = [
            operand
            for contract in self.contracts
            for operand in self.contract_metrics[contract.name].all_operands
        ]
        self.contract_metrics["ALL CONTRACTS"] = HalsteadContractMetrics(
            None, all_operators=all_operators, all_operands=all_operands
        )

    def update_reporting_sections(self) -> None:
        # Create the table and text for each section.
        data = {
            contract.name: self.contract_metrics[contract.name].to_dict()
            for contract in self.contracts
        }
        for title, attr, keys in self.SECTIONS:
            pretty_table = make_pretty_table(["Contract", *keys], data, False)
            section_title = f"{self.title} ({title})"
            txt = f"\n\n{section_title}:\n{pretty_table}\n"
            self.full_text += txt
            setattr(
                self,
                attr,
                SectionInfo(title=section_title, pretty_table=pretty_table, txt=txt),
            )



"""
Module detecting usage of more than one dynamic type in abi.encodePacked() arguments which could lead to collision
"""

from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Contract, SolidityFunction
from slither.core.variables import Variable
from slither.analyses.data_dependency.data_dependency import is_tainted
from slither.core.solidity_types import ElementaryType
from slither.core.solidity_types import ArrayType


def _is_dynamic_type(arg: Variable):
    """
    Args:
        arg (function argument)
    Returns:
        Bool
    """
    if isinstance(arg.type, ElementaryType) and (arg.type.name in ["string", "bytes"]):
        return True
    if isinstance(arg.type, ArrayType) and arg.type.length is None:
        return True

    return False


def _detect_abi_encodePacked_collision(contract: Contract):
    """
    Args:
        contract (Contract)
    Returns:
        list((Function), (list (Node)))
    """
    ret = []

    for f in contract.functions_and_modifiers_declared:
        for ir in f.solidity_calls:
            if ir.function == SolidityFunction("abi.encodePacked()"):
                dynamic_type_count = 0
                for arg in ir.arguments:
                    if is_tainted(arg, contract) and _is_dynamic_type(arg):
                        dynamic_type_count += 1
                    elif dynamic_type_count > 1:
                        ret.append((f, ir.node))
                        dynamic_type_count = 0
                    else:
                        dynamic_type_count = 0
                if dynamic_type_count > 1:
                    ret.append((f, ir.node))
    return ret


class EncodePackedCollision(AbstractDetector):
    """
    Detect usage of more than one dynamic type in abi.encodePacked() arguments which could to collision
    """

    ARGUMENT = "encode-packed-collision"
    HELP = "ABI encodePacked Collision"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#abi-encodePacked-collision"
    )

    WIKI_TITLE = "ABI encodePacked Collision"
    WIKI_DESCRIPTION = """Detect collision due to dynamic type usages in `abi.encodePacked`"""

    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Sign {
    function get_hash_for_signature(string name, string doc) external returns(bytes32) {
        return keccak256(abi.encodePacked(name, doc));
    }
}
```
Bob calls `get_hash_for_signature` with (`bob`, `This is the content`). The hash returned is used as an ID.
Eve creates a collision with the ID using (`bo`, `bThis is the content`) and compromises the system.
"""
    WIKI_RECOMMENDATION = """Do not use more than one dynamic type in `abi.encodePacked()`
(see the [Solidity documentation](https://solidity.readthedocs.io/en/v0.5.10/abi-spec.html?highlight=abi.encodePacked#non-standard-packed-modeDynamic)).
Use `abi.encode()`, preferably."""

    def _detect(self):
        """Detect usage of more than one dynamic type in abi.encodePacked(..) arguments which could lead to collision"""
        results = []
        for c in self.compilation_unit.contracts:
            values = _detect_abi_encodePacked_collision(c)
            for func, node in values:
                info = [
                    func,
                    " calls abi.encodePacked() with multiple dynamic arguments:\n\t- ",
                    node,
                    "\n",
                ]
                json = self.generate_result(info)
                results.append(json)

        return results



from typing import Any
from slither.slithir.operations.call import Call
from slither.slithir.variables.constant import Constant


class EventCall(Call):
    def __init__(self, name: str | Constant) -> None:
        super().__init__()
        self._name = name
        # todo add instance of the Event

    @property
    def name(self) -> str | Constant:
        return self._name

    @property
    def read(self) -> list[Any]:
        return self._unroll(self.arguments)

    def __str__(self):
        args = [str(a) for a in self.arguments]
        return f"Emit {self.name}({','.join(args)})"



from slither.slither import Slither

slither = Slither("find.sol")
find = slither.get_contract_from_name("Find")[0]

assert find

# Get the variable
my_variable = find.get_state_variable_from_name("my_variable")
assert my_variable


function_using_a_as_condition = [
    f
    for f in find.functions
    if f.is_reading_in_conditional_node(my_variable)
    or f.is_reading_in_require_or_assert(my_variable)
]

# Print the result
print(f'The function using "a" in condition are {[f.name for f in function_using_a_as_condition]}')



from dataclasses import dataclass, field
from pathlib import Path

from slither import Slither
from slither.utils.myprettytable import MyPrettyTable
from slither.utils.tests_pattern import is_test_file


@dataclass(slots=True)
class LoCInfo:
    loc: int = 0
    sloc: int = 0
    cloc: int = 0

    def total(self) -> int:
        return self.loc + self.sloc + self.cloc


@dataclass(slots=True)
class LoC:
    src: LoCInfo = field(default_factory=LoCInfo)
    dep: LoCInfo = field(default_factory=LoCInfo)
    test: LoCInfo = field(default_factory=LoCInfo)

    def to_pretty_table(self) -> MyPrettyTable:
        table = MyPrettyTable(["", "src", "dep", "test"])

        table.add_row(["loc", str(self.src.loc), str(self.dep.loc), str(self.test.loc)])
        table.add_row(["sloc", str(self.src.sloc), str(self.dep.sloc), str(self.test.sloc)])
        table.add_row(["cloc", str(self.src.cloc), str(self.dep.cloc), str(self.test.cloc)])
        table.add_row(
            ["Total", str(self.src.total()), str(self.dep.total()), str(self.test.total())]
        )
        return table


def count_lines(contract_lines: list[str]) -> tuple[int, int, int]:
    """Function to count and classify the lines of code in a contract.
    Args:
        contract_lines: list(str) representing the lines of a contract.
    Returns:
        tuple(int, int, int) representing (cloc, sloc, loc)
    """
    multiline_comment = False
    cloc = 0
    sloc = 0
    loc = 0

    for line in contract_lines:
        loc += 1
        stripped_line = line.strip()
        if not multiline_comment:
            if stripped_line.startswith("//"):
                cloc += 1
            elif "/*" in stripped_line:
                # Account for case where /* is followed by */ on the same line.
                # If it is, then multiline_comment does not need to be set to True
                start_idx = stripped_line.find("/*")
                end_idx = stripped_line.find("*/", start_idx + 2)
                if end_idx == -1:
                    multiline_comment = True
                cloc += 1
            elif stripped_line:
                sloc += 1
        else:
            cloc += 1
            if "*/" in stripped_line:
                multiline_comment = False

    return cloc, sloc, loc


def _update_lines(loc_info: LoCInfo, lines: list) -> None:
    """An internal function used to update (mutate in place) the loc_info.

    Args:
        loc_info: LoCInfo to be updated
        lines: list(str) representing the lines of a contract.
    """
    cloc, sloc, loc = count_lines(lines)
    loc_info.loc += loc
    loc_info.cloc += cloc
    loc_info.sloc += sloc


def compute_loc_metrics(slither: Slither) -> LoC:
    """Used to compute the lines of code metrics for a Slither object.

    Args:
        slither: A Slither object
    Returns:
        A LoC object
    """

    loc = LoC()

    for filename, source_code in slither.source_code.items():
        current_lines = source_code.splitlines()
        is_dep = False
        if slither.crytic_compile:
            is_dep = slither.crytic_compile.is_dependency(filename)
        loc_type = loc.dep if is_dep else loc.test if is_test_file(Path(filename)) else loc.src
        _update_lines(loc_type, current_lines)
    return loc



"""
Function module
"""

from typing import TYPE_CHECKING

from slither.core.declarations import Function
from slither.core.declarations.top_level import TopLevel
from slither.utils.using_for import USING_FOR, merge_using_for

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.core.scope.scope import FileScope
    from slither.slithir.variables.state_variable import StateIRVariable


class FunctionTopLevel(Function, TopLevel):
    def __init__(self, compilation_unit: "SlitherCompilationUnit", scope: "FileScope") -> None:
        super().__init__(compilation_unit)
        self._scope: FileScope = scope
        self._using_for_complete: USING_FOR | None = None

    @property
    def file_scope(self) -> "FileScope":
        return self._scope

    @property
    def using_for_complete(self) -> USING_FOR:
        """
        USING_FOR: Dict of top level directive
        """

        if self._using_for_complete is None:
            result = {}
            for uftl in self.file_scope.using_for_directives:
                result = merge_using_for(result, uftl.using_for)
            self._using_for_complete = result
        return self._using_for_complete

    @property
    def canonical_name(self) -> str:
        """
        str: contract.func_name(type1,type2)
        Return the function signature without the return values
        """
        if self._canonical_name is None:
            name, parameters, _ = self.signature
            self._canonical_name = (
                ".".join(self._internal_scope + [name]) + "(" + ",".join(parameters) + ")"
            )
        return self._canonical_name

    # endregion
    ###################################################################################
    ###################################################################################
    # region Functions
    ###################################################################################
    ###################################################################################

    @property
    def functions_shadowed(self) -> list["Function"]:
        return []

    # endregion
    ###################################################################################
    ###################################################################################
    # region Summary information
    ###################################################################################
    ###################################################################################

    def get_summary(
        self,
    ) -> tuple[str, str, str, list[str], list[str], list[str], list[str], list[str]]:
        """
            Return the function summary
        Returns:
            (str, str, str, list(str), list(str), listr(str), list(str), list(str);
            contract_name, name, visibility, modifiers, vars read, vars written, internal_calls, external_calls_as_expressions
        """
        return (
            "",
            self.full_name,
            self.visibility,
            [str(x) for x in self.modifiers],
            [str(x) for x in self.state_variables_read + self.solidity_variables_read],
            [str(x) for x in self.state_variables_written],
            [str(x) for x in self.internal_calls],
            [str(x) for x in self.external_calls_as_expressions],
        )

    # endregion
    ###################################################################################
    ###################################################################################
    # region SlithIr and SSA
    ###################################################################################
    ###################################################################################

    def generate_slithir_ssa(
        self, all_ssa_state_variables_instances: dict[str, "StateIRVariable"]
    ) -> None:
        from slither.slithir.utils.ssa import add_ssa_ir, transform_slithir_vars_to_ssa
        from slither.core.dominators.utils import (
            compute_dominance_frontier,
            compute_dominators,
        )

        compute_dominators(self.nodes)
        compute_dominance_frontier(self.nodes)
        transform_slithir_vars_to_ssa(self)

        add_ssa_ir(self, all_ssa_state_variables_instances)



import os
import re
from dataclasses import dataclass, field
from slither.core.declarations import Contract
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.vyper_parsing.declarations.contract import ContractVyper
from slither.analyses.data_dependency.data_dependency import compute_dependency
from slither.vyper_parsing.ast.types import Module
from slither.exceptions import SlitherException


@dataclass
class VyperCompilationUnit:
    _compilation_unit: SlitherCompilationUnit
    _parsed: bool = False
    _analyzed: bool = False
    _underlying_contract_to_parser: dict[Contract, ContractVyper] = field(default_factory=dict)
    _contracts_by_id: dict[int, Contract] = field(default_factory=dict)

    def parse_module(self, data: Module, filename: str):
        sourceUnit_candidates = re.findall("[0-9]*:[0-9]*:([0-9]*)", data.src)
        assert len(sourceUnit_candidates) == 1, "Source unit not found"
        sourceUnit = int(sourceUnit_candidates[0])

        self._compilation_unit.source_units[sourceUnit] = filename
        if os.path.isfile(filename) and filename not in self._compilation_unit.core.source_code:
            self._compilation_unit.core.add_source_code(filename)

        scope = self._compilation_unit.get_scope(filename)
        contract = Contract(self._compilation_unit, scope)
        contract_parser = ContractVyper(self, contract, data)
        contract.set_offset(data.src, self._compilation_unit)

        self._underlying_contract_to_parser[contract] = contract_parser

    def parse_contracts(self):
        for contract, contract_parser in self._underlying_contract_to_parser.items():
            self._contracts_by_id[contract.id] = contract
            self._compilation_unit.contracts.append(contract)

            contract_parser.parse_enums()
            contract_parser.parse_structs()
            contract_parser.parse_state_variables()
            contract_parser.parse_events()
            contract_parser.parse_functions()

        self._parsed = True

    def analyze_contracts(self) -> None:
        if not self._parsed:
            raise SlitherException("Parse the contract before running analyses")

        for contract_parser in self._underlying_contract_to_parser.values():
            # State variables are analyzed for all contracts because interfaces may
            # reference them, specifically, constants.
            contract_parser.analyze_state_variables()

        for contract_parser in self._underlying_contract_to_parser.values():
            contract_parser.analyze()

        self._convert_to_slithir()

        compute_dependency(self._compilation_unit)

        self._analyzed = True

    def _convert_to_slithir(self) -> None:
        for contract in self._compilation_unit.contracts:
            contract.add_constructor_variables()
            for func in contract.functions:
                func.generate_slithir_and_analyze()

            contract.convert_expression_to_slithir_ssa()

        self._compilation_unit.propagate_function_calls()
        for contract in self._compilation_unit.contracts:
            contract.fix_phi()
            contract.update_read_write_using_ssa()



"""
Module detecting unused state variables
"""

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import Function
from slither.core.declarations.contract import Contract
from slither.core.solidity_types import ArrayType
from slither.core.variables import Variable
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.formatters.variables.unused_state_variables import custom_format
from slither.utils.output import Output
from slither.visitors.expression.export_values import ExportValues


def detect_unused(contract: Contract) -> list[StateVariable] | None:
    # Get all the variables read in all the functions and modifiers

    all_functions = [
        f
        for f in contract.all_functions_called + list(contract.modifiers)
        if isinstance(f, Function)
    ]
    variables_used = [x.state_variables_read for x in all_functions]
    variables_used += [
        x.state_variables_written for x in all_functions if not x.is_constructor_variables
    ]

    array_candidates_ = [x.variables for x in all_functions]
    array_candidates: list[Variable] = [i for sl in array_candidates_ for i in sl]
    array_candidates += contract.state_variables
    array_candidates = [
        x.type.length for x in array_candidates if isinstance(x.type, ArrayType) and x.type.length
    ]
    array_candidates = [ExportValues(x).result() for x in array_candidates]
    array_candidates = [i for sl in array_candidates for i in sl]
    array_candidates = [v for v in array_candidates if isinstance(v, StateVariable)]

    # Flat list
    variables_used = [item for sublist in variables_used for item in sublist]
    variables_used = list(set(variables_used + array_candidates))

    # If the contract is abstract, only consider private variables as other visibilities may be used in dependencies
    if contract.is_abstract:
        return [
            x
            for x in contract.state_variables
            if x not in variables_used and x.visibility == "private"
        ]

    # Return the variables unused that are not public
    return [
        x for x in contract.state_variables if x not in variables_used and x.visibility != "public"
    ]


class UnusedStateVars(AbstractDetector):
    """
    Unused state variables detector
    """

    ARGUMENT = "unused-state"
    HELP = "Unused state variables"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unused-state-variable"

    WIKI_TITLE = "Unused state variable"
    WIKI_DESCRIPTION = "Unused state variable."
    WIKI_EXPLOIT_SCENARIO = ""
    WIKI_RECOMMENDATION = "Remove unused state variables."

    def _detect(self) -> list[Output]:
        """Detect unused state variables"""
        results = []
        for c in self.compilation_unit.contracts_derived:
            if c.is_signature_only() or c.is_library:
                continue
            unusedVars = detect_unused(c)
            if unusedVars:
                for var in unusedVars:
                    info: DETECTOR_INFO = [var, " is never used in ", c, "\n"]
                    json = self.generate_result(info)
                    results.append(json)

        return results

    @staticmethod
    def _format(compilation_unit: SlitherCompilationUnit, result: dict) -> None:
        custom_format(compilation_unit, result)



"""
Module detecting dangerous conversion to enum
"""

from slither.core.cfg.node import Node
from slither.core.declarations import Contract
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    make_solc_versions,
    DETECTOR_INFO,
)
from slither.slithir.operations import TypeConversion
from slither.core.declarations.enum import Enum
from slither.utils.output import Output


def _detect_dangerous_enum_conversions(contract: Contract) -> list[tuple[Node, SourceMapping]]:
    """Detect dangerous conversion to enum by checking IR
    Args:
         contract (Contract)
    Returns:
         list[(node, variable)] (Nodes where a variable is being converted into enum)
    """
    return [
        (n, ir.variable)
        for f in contract.functions_declared
        for n in f.nodes
        for ir in n.irs
        if isinstance(ir, TypeConversion) and isinstance(ir.type.type, Enum)
    ]


class EnumConversion(AbstractDetector):
    """
    Detect dangerous conversion to enum
    """

    ARGUMENT = "enum-conversion"
    HELP = "Detect dangerous enum conversion"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#dangerous-enum-conversion"
    WIKI_TITLE = "Dangerous enum conversion"
    WIKI_DESCRIPTION = "Detect out-of-range `enum` conversion (`solc` < `0.4.5`)."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
    pragma solidity 0.4.2;
    contract Test{

    enum E{a}

    function bug(uint a) public returns(E){
        return E(a);
    }
}
```
Attackers can trigger unexpected behaviour by calling `bug(1)`."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Use a recent compiler version. If `solc` <`0.4.5` is required, check the `enum` conversion range."

    VULNERABLE_SOLC_VERSIONS = make_solc_versions(4, 0, 4)

    def _detect(self) -> list[Output]:
        """Detect dangerous conversion to enum"""
        results = []

        for c in self.compilation_unit.contracts:
            ret = _detect_dangerous_enum_conversions(c)
            for node, var in ret:
                func_info: DETECTOR_INFO = [node, " has a dangerous enum conversion\n"]
                # Output each node with the function info header as a separate result.
                variable_info: DETECTOR_INFO = [
                    "\t- Variable: ",
                    var,
                    f" of type: {var.type!s}\n",
                ]
                node_info: DETECTOR_INFO = ["\t- Enum conversion: ", node, "\n"]
                json = self.generate_result(func_info + variable_info + node_info)
                results.append(json)

        return results



"""
Module detecting suicidal contract

A suicidal contract is an unprotected function that calls selfdestruct
"""

from slither.core.declarations import Function, Contract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class ProtectedVariables(AbstractDetector):
    ARGUMENT = "protected-vars"
    HELP = "Detected unprotected variables"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#protected-variables"

    WIKI_TITLE = "Protected Variables"
    WIKI_DESCRIPTION = "Detect unprotected variable that are marked protected"

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Buggy{

    /// @custom:security write-protection="onlyOwner()"
    address owner;

    function set_protected() public onlyOwner(){
        owner = msg.sender;
    }

    function set_not_protected() public{
        owner = msg.sender;
    }
}
```
`owner` must be always written by function using `onlyOwner` (`write-protection="onlyOwner()"`), however anyone can call `set_not_protected`.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Add access controls to the vulnerable function"

    def _analyze_function(self, function: Function, contract: Contract) -> list[Output]:
        results = []

        for state_variable_written in function.all_state_variables_written():
            if state_variable_written.write_protection:
                for function_sig in state_variable_written.write_protection:
                    function_protection = contract.get_function_from_signature(function_sig)
                    if not function_protection:
                        function_protection = contract.get_modifier_from_signature(function_sig)
                    if not function_protection:
                        self.logger.error(f"{function_sig} not found")
                        continue
                    if function_protection not in [
                        ir.function for ir in function.all_internal_calls()
                    ]:
                        info: DETECTOR_INFO = [
                            function,
                            " should have ",
                            function_protection,
                            " to protect ",
                            state_variable_written,
                            "\n",
                        ]

                        res = self.generate_result(info)
                        results.append(res)
        return results

    def _detect(self) -> list[Output]:
        """Detect the suicidal functions"""
        results = []
        for contract in self.compilation_unit.contracts_derived:
            for function in contract.functions_entry_points:
                results += self._analyze_function(function, contract)

        return results



import re

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.formatters.exceptions import FormatError
from slither.formatters.utils.patches import create_patch


def custom_format(compilation_unit: SlitherCompilationUnit, result: dict) -> None:
    for file_scope in compilation_unit.scopes.values():
        elements = result["elements"]
        for element in elements:
            if element["type"] != "function":
                # Skip variable elements
                continue
            target_contract = file_scope.get_contract_from_name(
                element["type_specific_fields"]["parent"]["name"]
            )
            if target_contract:
                function = target_contract.get_function_from_full_name(
                    element["type_specific_fields"]["signature"]
                )
                if function:
                    _patch(
                        compilation_unit,
                        result,
                        element["source_mapping"]["filename_absolute"],
                        int(
                            function.parameters_src().source_mapping.start
                            + function.parameters_src().source_mapping.length
                        ),
                        int(function.returns_src().source_mapping.start),
                    )


def _patch(
    compilation_unit: SlitherCompilationUnit,
    result: dict,
    in_file: str,
    modify_loc_start: int,
    modify_loc_end: int,
) -> None:
    in_file_str = compilation_unit.core.source_code[in_file].encode("utf8")
    old_str_of_interest = in_file_str[modify_loc_start:modify_loc_end]
    # Find the keywords view|pure|constant and remove them
    m = re.search("(view|pure|constant)", old_str_of_interest.decode("utf8"))
    if m:
        create_patch(
            result,
            in_file,
            modify_loc_start + m.span()[0],
            modify_loc_start + m.span()[1],
            m.groups(0)[0],  # this is view|pure|constant
            "",
        )
    else:
        raise FormatError(
            "No view/pure/constant specifier exists. Regex failed to remove specifier!"
        )



from .state_variable import StateVariable
from .variable import Variable
from .local_variable_init_from_tuple import LocalVariableInitFromTuple
from .local_variable import LocalVariable
from .top_level_variable import TopLevelVariable
from .event_variable import EventVariable
from .function_type_variable import FunctionTypeVariable
from .structure_variable import StructureVariable



from typing import TYPE_CHECKING, Optional, Union

from slither.core.declarations.contract_level import ContractLevel
from slither.core.declarations.top_level import TopLevel
from slither.core.expressions.expression import Expression
from slither.core.variables.variable import Variable


if TYPE_CHECKING:
    from slither.core.solidity_types.type import Type
    from slither.core.declarations import Contract, SolidityVariable, SolidityFunction
    from slither.solc_parsing.yul.evm_functions import YulBuiltin


class Identifier(Expression):
    def __init__(
        self,
        value: Union[
            Variable,
            "TopLevel",
            "ContractLevel",
            "Contract",
            "SolidityVariable",
            "SolidityFunction",
            "YulBuiltin",
        ],
    ) -> None:
        super().__init__()

        from slither.core.declarations import Contract, SolidityVariable, SolidityFunction
        from slither.solc_parsing.yul.evm_functions import YulBuiltin

        assert isinstance(
            value,
            (
                Variable,
                TopLevel,
                ContractLevel,
                Contract,
                SolidityVariable,
                SolidityFunction,
                YulBuiltin,
            ),
        )

        self._value: (
            Variable
            | TopLevel
            | ContractLevel
            | Contract
            | SolidityVariable
            | SolidityFunction
            | YulBuiltin
        ) = value
        self._type: Type | None = None

    @property
    def type(self) -> Optional["Type"]:
        return self._type

    @type.setter
    def type(self, new_type: "Type") -> None:
        self._type = new_type

    @property
    def value(
        self,
    ) -> Union[
        Variable,
        "TopLevel",
        "ContractLevel",
        "Contract",
        "SolidityVariable",
        "SolidityFunction",
        "YulBuiltin",
    ]:
        return self._value

    def __str__(self) -> str:
        return str(self._value)

    @property
    def expression(self):
        return self



"""
Detects the passing of arrays located in memory to functions which expect to modify arrays via storage reference.
"""

from slither.core.declarations import Function
from slither.core.variables import Variable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.solidity_types.array_type import ArrayType
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.local_variable import LocalVariable
from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.utils.output import Output


class ArrayByReference(AbstractDetector):
    """
    Detects passing of arrays located in memory to functions which expect to modify arrays via storage reference.
    """

    ARGUMENT = "array-by-reference"
    HELP = "Modifying storage array by value"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#modifying-storage-array-by-value"

    WIKI_TITLE = "Modifying storage array by value"
    WIKI_DESCRIPTION = (
        "Detect arrays passed to a function that expects reference to a storage array"
    )

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Memory {
    uint[1] public x; // storage

    function f() public {
        f1(x); // update x
        f2(x); // do not update x
    }

    function f1(uint[1] storage arr) internal { // by reference
        arr[0] = 1;
    }

    function f2(uint[1] arr) internal { // by value
        arr[0] = 2;
    }
}
```

Bob calls `f()`. Bob assumes that at the end of the call `x[0]` is 2, but it is 1.
As a result, Bob's usage of the contract is incorrect."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Ensure the correct usage of `memory` and `storage` in the function parameters. Make all the locations explicit."

    @staticmethod
    def get_funcs_modifying_array_params(contracts: list[Contract]) -> set[FunctionContract]:
        """
        Obtains a set of functions which take arrays not located in storage as parameters, and writes to them.
        :param contracts: The collection of contracts to check functions in.
        :return: A set of functions which take an array not located in storage as a parameter and writes to it.
        """
        # Initialize our resulting set of functions which modify non-reference array parameters
        results = set()

        # Loop through all functions in all contracts.
        for contract in contracts:
            for function in contract.functions_declared:
                # Skip any constructor functions.
                if function.is_constructor:
                    continue

                # Determine if this function takes an array as a parameter and the location isn't storage.
                # If it has been written to, we know this sets an non-storage-ref array.
                for param in function.parameters:
                    if isinstance(param.type, ArrayType) and param.location != "storage":
                        if param in function.variables_written:
                            results.add(function)
                            break

        return results

    @staticmethod
    def detect_calls_passing_ref_to_function(
        contracts: list[Contract], array_modifying_funcs: set[FunctionContract]
    ) -> list[tuple[Node, Variable, Function | Variable]]:
        """
        Obtains all calls passing storage arrays by value to a function which cannot write to them successfully.
        :param contracts: The collection of contracts to check for problematic calls in.
        :param array_modifying_funcs: The collection of functions which take non-storage arrays as input and writes to
        them.
        :return: A list of tuples (calling_node, affected_argument, invoked_function) which denote all problematic
        nodes invoking a function with some storage array argument where the invoked function seemingly attempts to
        write to the array unsuccessfully.
        """
        # Define our resulting array.
        results: list[tuple[Node, Variable, Function | Variable]] = []

        # Verify we have functions in our list to check for.
        if not array_modifying_funcs:
            return results

        # Loop for each node in each function/modifier in each contract

        for contract in contracts:
            for function in contract.functions_and_modifiers_declared:
                for ir in [ir for _, ir in function.high_level_calls] + function.internal_calls:
                    # Verify this references a function in our array modifying functions collection.
                    if ir.function not in array_modifying_funcs:
                        continue

                    # Verify one of these parameters is an array in storage.
                    for param, arg in zip(ir.function.parameters, ir.arguments, strict=False):
                        # Verify this argument is a variable that is an array type.
                        if not isinstance(arg, (StateVariable, LocalVariable)):
                            continue
                        if not isinstance(arg.type, ArrayType):
                            continue

                        # If it is a state variable OR a local variable referencing storage, we add it to the list.
                        if (
                            isinstance(arg, StateVariable)
                            or (isinstance(arg, LocalVariable) and arg.location == "storage")
                        ) and (isinstance(param.type, ArrayType) and param.location != "storage"):
                            results.append((ir.node, arg, ir.function))
        return results

    def _detect(self) -> list[Output]:
        """
        Detects passing of arrays located in memory to functions which expect to modify arrays via storage reference.
        :return: The JSON results of the detector, which contains the calling_node, affected_argument_variable and
        invoked_function for each result found.
        """
        results = []
        array_modifying_funcs = self.get_funcs_modifying_array_params(self.contracts)
        problematic_calls = self.detect_calls_passing_ref_to_function(
            self.contracts, array_modifying_funcs
        )

        if problematic_calls:
            for calling_node, affected_argument, invoked_function in problematic_calls:
                info: DETECTOR_INFO = [
                    calling_node.function,
                    " passes array ",
                    affected_argument,
                    " by reference to ",
                    invoked_function,
                    " which only takes arrays by value\n",
                ]

                res = self.generate_result(info)
                results.append(res)

        return results



"""
Module printing summary of the contract
"""

from slither.core.declarations import Function
from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.myprettytable import MyPrettyTable


class Modifiers(AbstractPrinter):
    ARGUMENT = "modifiers"
    HELP = "Print the modifiers called by each function"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#modifiers"

    def output(self, _filename):
        """
        _filename is not used
        Args:
            _filename(string)
        """

        all_txt = ""
        all_tables = []

        for contract in self.slither.contracts_derived:
            txt = f"\nContract {contract.name}"
            table = MyPrettyTable(["Function", "Modifiers"])
            for function in contract.functions:
                modifiers = function.modifiers
                for ir in function.all_internal_calls():
                    if isinstance(ir.function, Function):
                        modifiers += ir.function.modifiers
                for ir in function.all_library_calls():
                    if isinstance(ir.function, Function):
                        modifiers += ir.function.modifiers
                table.add_row([function.name, sorted([m.name for m in set(modifiers)])])
            txt += "\n" + str(table)
            self.info(txt)
            all_txt += txt
            all_tables.append((contract.name, table))

        res = self.generate_output(all_txt)
        for name, table in all_tables:
            res.add_pretty_table(table, name)

        return res



from slither.core.expressions.expression import Expression

from slither.core.solidity_types.type import Type


class MemberAccess(Expression):
    def __init__(self, member_name: str, member_type: str, expression: Expression) -> None:
        # assert isinstance(member_type, Type)
        # TODO member_type is not always a Type
        assert isinstance(expression, Expression)
        super().__init__()
        self._type: Type = member_type
        self._member_name: str = member_name
        self._expression: Expression = expression

    @property
    def expression(self) -> Expression:
        return self._expression

    @property
    def member_name(self) -> str:
        return self._member_name

    @property
    def type(self) -> Type:
        return self._type

    def __str__(self) -> str:
        return str(self.expression) + "." + self.member_name



import os
import argparse
from slither import Slither


def resolve_function(contract_name, function_name):
    """
    Resolves a function instance, given a contract name and function.
    :param contract_name: The name of the contract the function is declared in.
    :param function_name: The name of the function to resolve.
    :return: Returns the resolved function, raises an exception otherwise.
    """
    # Obtain the target contract
    contracts = slither.get_contract_from_name(contract_name)

    # Verify the contract was resolved successfully
    if len(contracts) != 1:
        raise ValueError(f"Could not resolve target contract: {contract_name}")
    contract = contracts[0]
    # Obtain the target function
    target_function = next(
        (function for function in contract.functions_declared if function.name == function_name),
        None,
    )

    # Verify we have resolved the function specified.
    if target_function is None:
        raise ValueError(f"Could not resolve target function: {contract_name}.{function_name}")

    # Add the resolved function to the new list.
    return target_function


def resolve_functions(functions):
    """
    Resolves the provided function descriptors.
    :param functions: A list of tuples (contract_name, function_name) or str (of form "ContractName.FunctionName")
    to resolve into function objects.
    :return: Returns a list of resolved functions.
    """
    # Create the resolved list.
    resolved = []

    # Verify that the provided argument is a list.
    if not isinstance(functions, list):
        raise ValueError("Provided functions to resolve must be a list type.")

    # Loop for each item in the list.
    for item in functions:
        if isinstance(item, str):
            # If the item is a single string, we assume it is of form 'ContractName.FunctionName'.
            parts = item.split(".")
            if len(parts) < 2:
                raise ValueError(
                    "Provided string descriptor must be of form 'ContractName.FunctionName'"
                )
            resolved.append(resolve_function(parts[0], parts[1]))
        elif isinstance(item, tuple):
            # If the item is a tuple, it should be a 2-tuple providing contract and function names.
            if len(item) != 2:
                raise ValueError(
                    "Provided tuple descriptor must provide a contract and function name."
                )
            resolved.append(resolve_function(item[0], item[1]))
        else:
            raise ValueError(
                f"Unexpected function descriptor type to resolve in list: {type(item)}"
            )

    # Return the resolved list.
    return resolved


def all_function_definitions(function):
    """
    Obtains a list of representing this function and any base definitions
    :param function: The function to obtain all definitions at and beneath.
    :return: Returns a list composed of the provided function definition and any base definitions.
    """
    return [function] + [
        f
        for c in function.contract.inheritance
        for f in c.functions_and_modifiers_declared
        if f.full_name == function.full_name
    ]


def __find_target_paths(target_function, current_path=None):
    current_path = current_path if current_path else []
    # Create our results list
    results = set()

    # Add our current function to the path.
    current_path = [target_function] + current_path

    # Obtain this target function and any base definitions.
    all_target_functions = set(all_function_definitions(target_function))

    # Look through all functions
    for contract in slither.contracts:
        for function in contract.functions_and_modifiers_declared:
            # If the function is already in our path, skip it.
            if function in current_path:
                continue

            # Find all function calls in this function (except for low level)
            called_functions = [f for (_, f) in function.high_level_calls + function.library_calls]
            called_functions += function.internal_calls
            called_functions = set(called_functions)

            # If any of our target functions are reachable from this function, it's a result.
            if all_target_functions.intersection(called_functions):
                path_results = __find_target_paths(function, current_path.copy())
                if path_results:
                    results = results.union(path_results)

    # If this path is external accessible from this point, we add the current path to the list.
    if target_function.visibility in ["public", "external"] and len(current_path) > 1:
        results.add(tuple(current_path))

    return results


def find_target_paths(target_functions):
    """
    Obtains all functions which can lead to any of the target functions being called.
    :param target_functions: The functions we are interested in reaching.
    :return: Returns a list of all functions which can reach any of the target_functions.
    """
    # Create our results list
    results = set()

    # Loop for each target function
    for target_function in target_functions:
        results = results.union(__find_target_paths(target_function))

    return results


def parse_args():
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="PossiblePaths",
        usage="possible_paths.py [--is-truffle] filename [contract.function targets]",
    )

    parser.add_argument(
        "--is-truffle",
        help="Indicates the filename refers to a truffle directory path.",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "filename", help="The filename of the contract or truffle directory to analyze."
    )

    parser.add_argument("targets", nargs="+")

    return parser.parse_args()


# ------------------------------
# PossiblePaths.py
#       Usage: python3 possible_paths.py [--is-truffle] filename targets
#       Example: python3 possible_paths.py contract.sol contract1.function1 contract2.function2 contract3.function3
# ------------------------------
# Parse all arguments
args = parse_args()

# If this is a truffle project, verify we have a valid build directory.
if args.is_truffle:
    cwd = os.path.abspath(args.filename)
    build_dir = os.path.join(cwd, "build", "contracts")
    if not os.path.exists(build_dir):
        raise FileNotFoundError(f"Could not find truffle build directory at '{build_dir}'")

# Perform slither analysis on the given filename
slither = Slither(args.filename, is_truffle=args.is_truffle)

targets = resolve_functions(args.targets)

# Print out all target functions.
print("Target functions:")
for target in targets:
    print(f"-{target.contract.name}.{target.full_name}")
print("\n")

# Obtain all paths which reach the target functions.
reaching_paths = find_target_paths(targets)
reaching_functions = {y for x in reaching_paths for y in x if y not in targets}

# Print out all function names which can reach the targets.
print("The following functions reach the specified targets:")
for function_desc in sorted([f"{f.canonical_name}" for f in reaching_functions]):
    print(f"-{function_desc}")
print("\n")

# Format all function paths.
reaching_paths_str = [
    " -> ".join([f"{f.canonical_name}" for f in reaching_path]) for reaching_path in reaching_paths
]

# Print a sorted list of all function paths which can reach the targets.
print("The following paths reach the specified targets:")
for reaching_path in sorted(reaching_paths_str):
    print(f"{reaching_path}\n")



from slither.core.cfg.node import Node
from slither.core.cfg.node import NodeType
from slither.core.expressions.assignment_operation import (
    AssignmentOperation,
    AssignmentOperationType,
)
from slither.core.expressions.identifier import Identifier
from slither.vyper_parsing.expressions.expression_parsing import parse_expression
from slither.visitors.expression.find_calls import FindCalls
from slither.visitors.expression.read_var import ReadVar
from slither.visitors.expression.write_var import WriteVar


class NodeVyper:
    def __init__(self, node: Node) -> None:
        self._unparsed_expression: dict | None = None
        self._node = node

    @property
    def underlying_node(self) -> Node:
        return self._node

    def add_unparsed_expression(self, expression: dict) -> None:
        assert self._unparsed_expression is None
        self._unparsed_expression = expression

    def analyze_expressions(self, caller_context) -> None:
        if self._node.type == NodeType.VARIABLE and not self._node.expression:
            self._node.add_expression(self._node.variable_declaration.expression)
        if self._unparsed_expression:
            expression = parse_expression(self._unparsed_expression, caller_context)
            self._node.add_expression(expression)
            self._unparsed_expression = None

        if self._node.expression:
            if self._node.type == NodeType.VARIABLE:
                # Update the expression to be an assignment to the variable
                _expression = AssignmentOperation(
                    Identifier(self._node.variable_declaration),
                    self._node.expression,
                    AssignmentOperationType.ASSIGN,
                    self._node.variable_declaration.type,
                )
                _expression.set_offset(
                    self._node.expression.source_mapping, self._node.compilation_unit
                )
                self._node.add_expression(_expression, bypass_verif_empty=True)

            expression = self._node.expression
            read_var = ReadVar(expression)
            self._node.variables_read_as_expression = read_var.result()

            write_var = WriteVar(expression)
            self._node.variables_written_as_expression = write_var.result()

            find_call = FindCalls(expression)
            self._node.calls_as_expression = find_call.result()
            self._node.external_calls_as_expressions = [
                c for c in self._node.calls_as_expression if not isinstance(c.called, Identifier)
            ]
            self._node.internal_calls_as_expressions = [
                c for c in self._node.calls_as_expression if isinstance(c.called, Identifier)
            ]



from pathlib import Path

from slither import Slither
from slither.core.declarations.function import FunctionType

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_fallback_receive(solc_binary_path):
    solc_path = solc_binary_path("0.6.12")
    file = Path(TEST_DATA_DIR, "fallback.sol").as_posix()
    slither = Slither(file, solc=solc_path)
    fake_fallback = slither.get_contract_from_name("FakeFallback")[0]
    real_fallback = slither.get_contract_from_name("Fallback")[0]

    assert fake_fallback.fallback_function is None
    assert fake_fallback.receive_function is None
    assert real_fallback.fallback_function.function_type == FunctionType.FALLBACK
    assert real_fallback.receive_function.function_type == FunctionType.RECEIVE



import argparse
import enum
import json
import os
import re
import logging
from collections import defaultdict

from crytic_compile.cryticparser.defaults import (
    DEFAULTS_FLAG_IN_CONFIG as DEFAULTS_FLAG_IN_CONFIG_CRYTIC_COMPILE,
)

from slither.detectors.abstract_detector import classification_txt, AbstractDetector
from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.colors import yellow, red
from slither.utils.myprettytable import MyPrettyTable

logger = logging.getLogger("Slither")

DEFAULT_JSON_OUTPUT_TYPES = ["detectors", "printers"]
JSON_OUTPUT_TYPES = [
    "compilations",
    "console",
    "detectors",
    "printers",
    "list-detectors",
    "list-printers",
]


class FailOnLevel(enum.Enum):
    PEDANTIC = "pedantic"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    NONE = "none"


# Those are the flags shared by the command line and the config file
defaults_flag_in_config = {
    "codex": False,
    "codex_contracts": "all",
    "codex_model": "text-davinci-003",
    "codex_temperature": 0,
    "codex_max_tokens": 300,
    "codex_log": False,
    "detectors_to_run": "all",
    "printers_to_run": None,
    "detectors_to_exclude": None,
    "detectors_to_include": None,
    "exclude_dependencies": False,
    "exclude_informational": False,
    "exclude_optimization": False,
    "exclude_low": False,
    "exclude_medium": False,
    "exclude_high": False,
    "fail_on": FailOnLevel.PEDANTIC,
    "json": None,
    "sarif": None,
    "json-types": ",".join(DEFAULT_JSON_OUTPUT_TYPES),
    "disable_color": False,
    "filter_paths": None,
    "include_paths": None,
    "generate_patches": False,
    # debug command
    "skip_assembly": False,
    "legacy_ast": False,
    "zip": None,
    "zip_type": "lzma",
    "show_ignored_findings": False,
    "no_fail": False,
    "sarif_input": "export.sarif",
    "sarif_triage": "export.sarif.sarifexplorer",
    "triage_database": "slither.db.json",
    **DEFAULTS_FLAG_IN_CONFIG_CRYTIC_COMPILE,
}


def read_config_file(args: argparse.Namespace) -> None:
    # No config file was provided as an argument
    if args.config_file is None:
        # Check whether the default config file is present
        if os.path.exists("slither.config.json"):
            # The default file exists, use it
            args.config_file = "slither.config.json"
        else:
            return

    if os.path.isfile(args.config_file):
        try:
            with open(args.config_file, encoding="utf8") as f:
                config = json.load(f)
                for key, elem in config.items():
                    if key not in defaults_flag_in_config:
                        logger.info(
                            yellow(f"{args.config_file} has an unknown key: {key} : {elem}")
                        )
                        continue
                    if getattr(args, key) == defaults_flag_in_config[key]:
                        setattr(args, key, elem)
        except json.decoder.JSONDecodeError as e:
            logger.error(red(f"Impossible to read {args.config_file}, please check the file {e}"))
    else:
        logger.error(red(f"File {args.config_file} is not a file or does not exist"))
        logger.error(yellow("Falling back to the default settings..."))


def output_to_markdown(
    detector_classes: list[type[AbstractDetector]],
    printer_classes: list[type[AbstractPrinter]],
    filter_wiki: str,
) -> None:
    def extract_help(cls: type[AbstractDetector] | type[AbstractPrinter]) -> str:
        if cls.WIKI == "":
            return cls.HELP
        return f"[{cls.HELP}]({cls.WIKI})"

    detectors_list = []
    print(filter_wiki)
    for detector in detector_classes:
        argument = detector.ARGUMENT
        # dont show the backdoor example
        if argument == "backdoor":
            continue
        if filter_wiki not in detector.WIKI:
            continue
        help_info = extract_help(detector)
        impact = detector.IMPACT
        confidence = classification_txt[detector.CONFIDENCE]
        detectors_list.append((argument, help_info, impact, confidence))

    # Sort by impact, confidence, and name
    detectors_list = sorted(
        detectors_list, key=lambda element: (element[2], element[3], element[0])
    )
    idx = 1
    for argument, help_info, impact, confidence in detectors_list:
        print(f"{idx} | `{argument}` | {help_info} | {classification_txt[impact]} | {confidence}")
        idx = idx + 1

    print()
    printers_list = []
    for printer in printer_classes:
        argument = printer.ARGUMENT
        help_info = extract_help(printer)
        printers_list.append((argument, help_info))

    # Sort by impact, confidence, and name
    printers_list = sorted(printers_list, key=lambda element: (element[0]))
    idx = 1
    for argument, help_info in printers_list:
        print(f"{idx} | `{argument}` | {help_info}")
        idx = idx + 1


def get_level(l: str) -> int:
    tab = l.count("\t") + 1
    if l.replace("\t", "").startswith(" -"):
        tab = tab + 1
    if l.replace("\t", "").startswith("-"):
        tab = tab + 1
    return tab


def convert_result_to_markdown(txt: str) -> str:
    # -1 to remove the last \n
    lines = txt[0:-1].split("\n")
    ret = []
    level = 0
    for l in lines:
        next_level = get_level(l)
        prefix = "<li>"
        if next_level < level:
            prefix = "</ul>" * (level - next_level) + prefix
        if next_level > level:
            prefix = "<ul>" * (next_level - level) + prefix
        level = next_level
        ret.append(prefix + l)

    return "".join(ret)


def output_results_to_markdown(
    all_results: list[dict], checklistlimit: str, show_ignored_findings: bool
) -> None:
    checks = defaultdict(list)
    info: dict = defaultdict(dict)
    for results_ in all_results:
        checks[results_["check"]].append(results_)
        info[results_["check"]] = {
            "impact": results_["impact"],
            "confidence": results_["confidence"],
        }

    if not show_ignored_findings:
        print(
            "**THIS CHECKLIST IS NOT COMPLETE**. Use `--show-ignored-findings` to show all the results."
        )

    print("Summary")
    for check_ in checks:
        print(
            f" - [{check_}](#{check_}) ({len(checks[check_])} results) ({info[check_]['impact']})"
        )

    counter = 0
    for check, results in checks.items():
        print(f"## {check}")
        print(f"Impact: {info[check]['impact']}")
        print(f"Confidence: {info[check]['confidence']}")
        additional = False
        if checklistlimit and len(results) > 5:
            results = results[0:5]
            additional = True
        for result in results:
            print(" - [ ] ID-" + f"{counter}")
            counter = counter + 1
            print(result["markdown"])
            if result["first_markdown_element"]:
                print(result["first_markdown_element"])
                print("\n")
        if additional:
            print(f"**More results were found, check [{checklistlimit}]({checklistlimit})**")


def output_wiki(detector_classes: list[type[AbstractDetector]], filter_wiki: str) -> None:
    # Sort by impact, confidence, and name
    detectors_list = sorted(
        detector_classes,
        key=lambda element: (element.IMPACT, element.CONFIDENCE, element.ARGUMENT),
    )

    for detector in detectors_list:
        argument = detector.ARGUMENT
        # dont show the backdoor example
        if argument == "backdoor":
            continue
        if filter_wiki not in detector.WIKI:
            continue
        check = detector.ARGUMENT
        impact = classification_txt[detector.IMPACT]
        confidence = classification_txt[detector.CONFIDENCE]
        title = detector.WIKI_TITLE
        description = detector.WIKI_DESCRIPTION
        exploit_scenario = detector.WIKI_EXPLOIT_SCENARIO
        recommendation = detector.WIKI_RECOMMENDATION

        print(f"\n## {title}")
        print("### Configuration")
        print(f"* Check: `{check}`")
        print(f"* Severity: `{impact}`")
        print(f"* Confidence: `{confidence}`")
        print("\n### Description")
        print(description)
        if exploit_scenario:
            print("\n### Exploit Scenario:")
            print(exploit_scenario)
        print("\n### Recommendation")
        print(recommendation)


def output_detectors(detector_classes: list[type[AbstractDetector]]) -> None:
    detectors_list = []
    for detector in detector_classes:
        argument = detector.ARGUMENT
        # dont show the backdoor example
        if argument == "backdoor":
            continue
        help_info = detector.HELP
        impact = detector.IMPACT
        confidence = classification_txt[detector.CONFIDENCE]
        detectors_list.append((argument, help_info, impact, confidence))
    table = MyPrettyTable(["Num", "Check", "What it Detects", "Impact", "Confidence"])

    # Sort by impact, confidence, and name
    detectors_list = sorted(
        detectors_list, key=lambda element: (element[2], element[3], element[0])
    )
    idx = 1
    for argument, help_info, impact, confidence in detectors_list:
        table.add_row([str(idx), argument, help_info, classification_txt[impact], confidence])
        idx = idx + 1
    print(table)


def output_detectors_json(
    detector_classes: list[type[AbstractDetector]],
) -> list[dict]:
    detectors_list = []
    for detector in detector_classes:
        argument = detector.ARGUMENT
        # dont show the backdoor example
        if argument == "backdoor":
            continue
        help_info = detector.HELP
        impact = detector.IMPACT
        confidence = classification_txt[detector.CONFIDENCE]
        wiki_url = detector.WIKI
        wiki_description = detector.WIKI_DESCRIPTION
        wiki_exploit_scenario = detector.WIKI_EXPLOIT_SCENARIO
        wiki_recommendation = detector.WIKI_RECOMMENDATION
        detectors_list.append(
            (
                argument,
                help_info,
                impact,
                confidence,
                wiki_url,
                wiki_description,
                wiki_exploit_scenario,
                wiki_recommendation,
            )
        )

    # Sort by impact, confidence, and name
    detectors_list = sorted(
        detectors_list, key=lambda element: (element[2], element[3], element[0])
    )
    idx = 1
    table = []
    for (
        argument,
        help_info,
        impact,
        confidence,
        wiki_url,
        description,
        exploit,
        recommendation,
    ) in detectors_list:
        table.append(
            {
                "index": idx,
                "check": argument,
                "title": help_info,
                "impact": classification_txt[impact],
                "confidence": confidence,
                "wiki_url": wiki_url,
                "description": description,
                "exploit_scenario": exploit,
                "recommendation": recommendation,
            }
        )
        idx = idx + 1
    return table


def output_printers(printer_classes: list[type[AbstractPrinter]]) -> None:
    printers_list = []
    for printer in printer_classes:
        argument = printer.ARGUMENT
        help_info = printer.HELP
        printers_list.append((argument, help_info))
    table = MyPrettyTable(["Num", "Printer", "What it Does"])

    # Sort by impact, confidence, and name
    printers_list = sorted(printers_list, key=lambda element: (element[0]))
    idx = 1
    for argument, help_info in printers_list:
        # Clean multi line HELP info
        table.add_row([str(idx), argument, " ".join(x.strip() for x in help_info.splitlines())])
        idx = idx + 1

    print(table)


def output_printers_json(printer_classes: list[type[AbstractPrinter]]) -> list[dict]:
    printers_list = []
    for printer in printer_classes:
        argument = printer.ARGUMENT
        help_info = printer.HELP

        printers_list.append((argument, help_info))

    # Sort by name
    printers_list = sorted(printers_list, key=lambda element: (element[0]))
    idx = 1
    table = []
    for argument, help_info in printers_list:
        table.append({"index": idx, "check": argument, "title": help_info})
        idx = idx + 1
    return table


def check_and_sanitize_markdown_root(markdown_root: str) -> str:
    # Regex to check whether the markdown_root is a GitHub URL
    match = re.search(
        r"(https://)github.com/([a-zA-Z-]+)([:/][A-Za-z0-9_.-]+[:/]?)([A-Za-z0-9_.-]*)(.*)",
        markdown_root,
    )
    if match:
        if markdown_root[-1] != "/":
            logger.warning("Appending '/' in markdown_root url for better code referencing")
            markdown_root = markdown_root + "/"

        if not match.group(4):
            logger.warning(
                "Appending 'master/tree/' in markdown_root url for better code referencing"
            )
            markdown_root = markdown_root + "master/tree/"
        elif match.group(4) == "tree":
            logger.warning(
                "Replacing 'tree' with 'blob' in markdown_root url for better code referencing"
            )
            positions = match.span(4)
            markdown_root = f"{markdown_root[: positions[0]]}blob{markdown_root[positions[1] :]}"

    return markdown_root



import logging
from enum import Enum

from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue, is_valid_rvalue
from slither.slithir.exceptions import SlithIRError
from slither.core.variables.local_variable import LocalVariable
from slither.slithir.variables.constant import Constant
from slither.slithir.variables.local_variable import LocalIRVariable
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA


logger = logging.getLogger("BinaryOperationIR")


class UnaryType(Enum):
    BANG = "!"
    TILD = "~"

    @staticmethod
    def get_type(operation_type, isprefix):
        if isprefix:
            if operation_type == "!":
                return UnaryType.BANG
            if operation_type == "~":
                return UnaryType.TILD
        raise SlithIRError(f"get_type: Unknown operation type {operation_type}")


class Unary(OperationWithLValue):
    def __init__(
        self,
        result: TemporaryVariableSSA | TemporaryVariable,
        variable: Constant | LocalIRVariable | LocalVariable,
        operation_type: UnaryType,
    ) -> None:
        assert is_valid_rvalue(variable)
        assert is_valid_lvalue(result)
        super().__init__()
        self._variable = variable
        self._type = operation_type
        self._lvalue = result

    @property
    def read(self) -> list[Constant | LocalIRVariable | LocalVariable]:
        return [self._variable]

    @property
    def rvalue(self) -> Constant | LocalVariable:
        return self._variable

    @property
    def type(self) -> UnaryType:
        return self._type

    @property
    def type_str(self):
        return str(self._type)

    def __str__(self):
        return f"{self.lvalue} = {self.type_str} {self.rvalue} "



from slither.tools.upgradeability.checks.initialization import (
    InitializablePresent,
    InitializableInherited,
    InitializableInitializer,
    MissingInitializerModifier,
    MissingCalls,
    MultipleCalls,
    InitializeTarget,
    MultipleReinitializers,
)

from slither.tools.upgradeability.checks.functions_ids import IDCollision, FunctionShadowing

from slither.tools.upgradeability.checks.variable_initialization import VariableWithInit

from slither.tools.upgradeability.checks.variables_order import (
    MissingVariable,
    DifferentVariableContractProxy,
    DifferentVariableContractNewContract,
    ExtraVariablesProxy,
    ExtraVariablesNewContract,
)

from slither.tools.upgradeability.checks.constant import WereConstant, BecameConstant



from typing import TYPE_CHECKING

from slither.core.declarations import Structure
from slither.core.declarations.top_level import TopLevel

if TYPE_CHECKING:
    from slither.core.scope.scope import FileScope
    from slither.core.compilation_unit import SlitherCompilationUnit


class StructureTopLevel(Structure, TopLevel):
    def __init__(self, compilation_unit: "SlitherCompilationUnit", scope: "FileScope") -> None:
        super().__init__(compilation_unit)
        self.file_scope: FileScope = scope



from typing import TYPE_CHECKING

from slither.core.declarations.top_level import TopLevel
from slither.core.declarations.contract_level import ContractLevel
from slither.core.solidity_types import Type, ElementaryType

if TYPE_CHECKING:
    from slither.core.declarations.function_top_level import FunctionTopLevel
    from slither.core.declarations import Contract
    from slither.core.scope.scope import FileScope


class TypeAlias(Type):
    def __init__(self, underlying_type: ElementaryType, name: str) -> None:
        super().__init__()
        self.name = name
        self.underlying_type = underlying_type
        self._pattern = "type"

    @property
    def type(self) -> ElementaryType:
        """
        Return the underlying type. Alias for underlying_type


        Returns:
            Type: the underlying type

        """
        return self.underlying_type

    @property
    def storage_size(self) -> tuple[int, bool]:
        return self.underlying_type.storage_size

    def __hash__(self) -> int:
        return hash(str(self))

    @property
    def is_dynamic(self) -> bool:
        return self.underlying_type.is_dynamic


class TypeAliasTopLevel(TypeAlias, TopLevel):
    def __init__(self, underlying_type: ElementaryType, name: str, scope: "FileScope") -> None:
        super().__init__(underlying_type, name)
        self.file_scope: FileScope = scope
        # operators redefined
        self.operators: dict[str, FunctionTopLevel] = {}

    def __str__(self) -> str:
        return self.name


class TypeAliasContract(TypeAlias, ContractLevel):
    def __init__(self, underlying_type: ElementaryType, name: str, contract: "Contract") -> None:
        super().__init__(underlying_type, name)
        self._contract: Contract = contract

    def __str__(self) -> str:
        return self.contract.name + "." + self.name



"""
Module detecting bad PRNG due to the use of block.timestamp, now or blockhash (block.blockhash) as a source of randomness
"""

from slither.analyses.data_dependency.data_dependency import is_dependent_ssa
from slither.core.cfg.node import Node
from slither.core.declarations import Function, Contract
from slither.core.declarations.solidity_variables import (
    SolidityVariable,
    SolidityFunction,
    SolidityVariableComposed,
)
from slither.core.variables.variable import Variable
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import BinaryType, Binary
from slither.slithir.operations import SolidityCall
from slither.utils.output import Output, AllSupportedOutput


def collect_return_values_of_bad_PRNG_functions(f: Function) -> list:
    """
        Return the return-values of calls to blockhash()
    Args:
        f (Function)
    Returns:
        list(values)
    """
    values_returned = []
    for n in f.nodes:
        for ir in n.irs_ssa:
            if (
                isinstance(ir, SolidityCall)
                and ir.function == SolidityFunction("blockhash(uint256)")
                and ir.lvalue
            ):
                values_returned.append(ir.lvalue)
    return values_returned


def contains_bad_PRNG_sources(func: Function, blockhash_ret_values: list[Variable]) -> list[Node]:
    """
         Check if any node in function has a modulus operator and the first operand is dependent on block.timestamp, now or blockhash()
    Returns:
        (nodes)
    """
    ret = set()

    for node in func.nodes:
        for ir in node.irs_ssa:
            if isinstance(ir, Binary) and ir.type == BinaryType.MODULO:
                var_left = ir.variable_left
                if not isinstance(var_left, (Variable, SolidityVariable)):
                    continue
                if is_dependent_ssa(
                    var_left, SolidityVariableComposed("block.timestamp"), node
                ) or is_dependent_ssa(var_left, SolidityVariable("now"), node):
                    ret.add(node)
                    break

                for ret_val in blockhash_ret_values:
                    if is_dependent_ssa(var_left, ret_val, node):
                        ret.add(node)
                        break
    return list(ret)


def detect_bad_PRNG(contract: Contract) -> list[tuple[Function, list[Node]]]:
    """
    Args:
        contract (Contract)
    Returns:
        list((Function), (list (Node)))
    """
    blockhash_ret_values = []
    for f in contract.functions:
        blockhash_ret_values += collect_return_values_of_bad_PRNG_functions(f)
    ret: list[tuple[Function, list[Node]]] = []
    for f in contract.functions:
        bad_prng_nodes = contains_bad_PRNG_sources(f, blockhash_ret_values)
        if bad_prng_nodes:
            ret.append((f, bad_prng_nodes))
    return ret


class BadPRNG(AbstractDetector):
    """
    Detect weak PRNG due to a modulo operation on block.timestamp, now or blockhash
    """

    ARGUMENT = "weak-prng"
    HELP = "Weak PRNG"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#weak-PRNG"

    WIKI_TITLE = "Weak PRNG"
    WIKI_DESCRIPTION = "Weak PRNG due to a modulo on `block.timestamp`, `now` or `blockhash`. These can be influenced by miners to some extent so they should be avoided."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Game {

    uint reward_determining_number;

    function guessing() external{
      reward_determining_number = uint256(block.blockhash(10000)) % 10;
    }
}
```
Eve is a miner. Eve calls `guessing` and re-orders the block containing the transaction.
As a result, Eve wins the game."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        "Do not use `block.timestamp`, `now` or `blockhash` as a source of randomness"
    )

    def _detect(self) -> list[Output]:
        """Detect bad PRNG due to the use of block.timestamp, now or blockhash (block.blockhash) as a source of randomness"""
        results = []
        for c in self.compilation_unit.contracts_derived:
            values = detect_bad_PRNG(c)
            for func, nodes in values:
                for node in nodes:
                    info: list[AllSupportedOutput] = [func, ' uses a weak PRNG: "', node, '" \n']
                    res = self.generate_result(info)
                    results.append(res)

        return results



import argparse
import logging
import uuid
from crytic_compile import cryticparser
from slither import Slither
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import Function

from slither.formatters.utils.patches import create_patch, apply_patch, create_diff
from slither.utils import codex

logging.basicConfig()
logging.getLogger("Slither").setLevel(logging.INFO)

logger = logging.getLogger("Slither")


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="Auto-generate NatSpec documentation for every function using OpenAI Codex.",
        usage="slither-documentation filename",
    )

    parser.add_argument("project", help="The target directory/Solidity file.")

    parser.add_argument(
        "--overwrite", help="Overwrite the files (be careful).", action="store_true", default=False
    )

    parser.add_argument(
        "--force-answer-parsing",
        help="Apply heuristics to better parse codex output (might lead to incorrect results)",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--include-tests",
        help="Include the tests",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--retry",
        help="Retry failed query (default 1). Each retry increases the temperature by 0.1",
        action="store",
        default=1,
    )

    # Add default arguments from crytic-compile
    cryticparser.init(parser)

    codex.init_parser(parser, always_enable_codex=True)

    return parser.parse_args()


def _use_tab(char: str) -> bool | None:
    """
    Check if the char is a tab

    Args:
        char:

    Returns:

    """
    if char == " ":
        return False
    if char == "\t":
        return True
    return None


def _post_processesing(
    answer: str, starting_column: int, use_tab: bool | None, force_and_stopped: bool
) -> str | None:
    """
    Clean answers from codex

    Args:
        answer:
        starting_column:

    Returns:

    """
    if answer.count("/**") != 1:
        return None
    # Sometimes codex will miss the */, even if it finished properly the request
    # In this case, we allow slither-documentation to force the */
    if answer.count("*/") != 1:
        if force_and_stopped:
            answer += "*/"
        else:
            return None
    if answer.find("/**") > answer.find("*/"):
        return None
    answer = answer[answer.find("/**") : answer.find("*/") + 2]
    answer_lines = answer.splitlines()
    # Add indentation to all the lines, aside the first one

    space_char = "\t" if use_tab else " "

    if len(answer_lines) > 0:
        answer = (
            answer_lines[0]
            + "\n"
            + "\n".join(
                [space_char * (starting_column - 1) + line for line in answer_lines[1:] if line]
            )
        )
        answer += "\n" + space_char * (starting_column - 1)
        return answer
    return answer_lines[0]


def _handle_codex(
    answer: dict, starting_column: int, use_tab: bool | None, force: bool
) -> str | None:
    if "choices" in answer:
        if answer["choices"]:
            if "text" in answer["choices"][0]:
                has_stopped = answer["choices"][0].get("finish_reason", "") == "stop"
                answer_processed = _post_processesing(
                    answer["choices"][0]["text"], starting_column, use_tab, force and has_stopped
                )
                if answer_processed is None:
                    return None
                return answer_processed
    return None


def _handle_function(
    function: Function,
    overwrite: bool,
    all_patches: dict,
    logging_file: str | None,
    slither: Slither,
    retry: int,
    force: bool,
) -> bool:
    if (
        function.source_mapping.is_dependency
        or function.has_documentation
        or function.is_constructor_variables
    ):
        return overwrite
    prompt = "Create a natpsec documentation for this solidity code with only notice and dev.\n"
    srcmap = function.source_mapping
    src = function.compilation_unit.core.source_code[srcmap.filename.absolute]
    first_char_index = len(
        src.encode("utf8")[: srcmap.start].decode("utf8")
    )  # convert byte offset to char offset
    prev_char = src[first_char_index - 1]
    prompt += srcmap.content

    use_tab = _use_tab(prev_char)
    if use_tab is None and srcmap.starting_column > 1:
        logger.info(f"Non standard indentation found: '{prev_char}'")
        if overwrite:
            logger.info("Disable overwrite to avoid mistakes")
            overwrite = False

    openai = codex.openai_module()  # type: ignore
    if openai is None:
        raise ImportError

    if logging_file:
        codex.log_codex(logging_file, "Q: " + prompt)

    tentative = 0
    answer_processed: str | None = None
    while tentative < retry:
        tentative += 1

        answer = openai.Completion.create(  # type: ignore
            prompt=prompt,
            model=slither.codex_model,
            temperature=min(slither.codex_temperature + tentative * 0.1, 1),
            max_tokens=slither.codex_max_tokens,
        )

        if logging_file:
            codex.log_codex(logging_file, "A: " + str(answer))

        answer_processed = _handle_codex(answer, srcmap.starting_column, use_tab, force)
        if answer_processed:
            break

        logger.info(
            f"Codex could not generate a well formatted answer for {function.canonical_name}"
        )
        logger.info(answer)

    if not answer_processed:
        return overwrite

    create_patch(
        all_patches, srcmap.filename.absolute, srcmap.start, srcmap.start, "", answer_processed
    )

    return overwrite


def _handle_compilation_unit(
    slither: Slither,
    compilation_unit: SlitherCompilationUnit,
    overwrite: bool,
    force: bool,
    retry: int,
    include_test: bool,
) -> None:
    logging_file: str | None
    if slither.codex_log:
        logging_file = str(uuid.uuid4())
    else:
        logging_file = None

    for scope in compilation_unit.scopes.values():
        # Dont send tests file
        if not include_test and (
            ".t.sol" in scope.filename.absolute
            or "mock" in scope.filename.absolute.lower()
            or "test" in scope.filename.absolute.lower()
        ):
            continue

        functions_target: list[Function] = []

        for contract in scope.contracts.values():
            functions_target += contract.functions_declared

        functions_target += list(scope.functions)

        all_patches: dict = {}

        for function in functions_target:
            overwrite = _handle_function(
                function, overwrite, all_patches, logging_file, slither, retry, force
            )

        # all_patches["patches"] should have only 1 file
        if "patches" not in all_patches:
            continue
        for file in all_patches["patches"]:
            original_txt = compilation_unit.core.source_code[file].encode("utf8")
            patched_txt = original_txt

            patches = all_patches["patches"][file]
            offset = 0
            patches.sort(key=lambda x: x["start"])

            for patch in patches:
                patched_txt, offset = apply_patch(patched_txt, patch, offset)

            if overwrite:
                with open(file, "w", encoding="utf8") as f:
                    f.write(patched_txt.decode("utf8"))
            else:
                diff = create_diff(compilation_unit, original_txt, patched_txt, file)
                with open(f"{file}.patch", "w", encoding="utf8") as f:
                    f.write(diff)


def main() -> None:
    args = parse_args()

    logger.info("This tool is a WIP, use it with cautious")
    logger.info("Be aware of OpenAI ToS: https://openai.com/api/policies/terms/")
    slither = Slither(args.project, **vars(args))

    try:
        for compilation_unit in slither.compilation_units:
            _handle_compilation_unit(
                slither,
                compilation_unit,
                args.overwrite,
                args.force_answer_parsing,
                int(args.retry),
                args.include_tests,
            )
    except ImportError:
        pass


if __name__ == "__main__":
    main()



import os
import difflib
from collections import defaultdict

from slither.core.compilation_unit import SlitherCompilationUnit


def create_patch(
    result: dict,
    file: str,
    start: int,
    end: int,
    old_str: str | bytes,
    new_str: str | bytes,
) -> None:
    if isinstance(old_str, bytes):
        old_str = old_str.decode("utf8")
    if isinstance(new_str, bytes):
        new_str = new_str.decode("utf8")
    p = {"start": start, "end": end, "old_string": old_str, "new_string": new_str}
    if "patches" not in result:
        result["patches"] = defaultdict(list)
    if p not in result["patches"][file]:
        result["patches"][file].append(p)


def apply_patch(original_txt: bytes, patch: dict, offset: int) -> tuple[bytes, int]:
    patched_txt = original_txt[: int(patch["start"] + offset)]
    patched_txt += patch["new_string"].encode("utf8")
    patched_txt += original_txt[int(patch["end"] + offset) :]

    # Keep the diff of text added or sub, in case of multiple patches
    # Note: must use byte length since patch offsets are byte offsets from solc
    patch_length_diff = len(patch["new_string"].encode("utf8")) - (patch["end"] - patch["start"])
    return patched_txt, patch_length_diff + offset


def create_diff(
    compilation_unit: SlitherCompilationUnit, original_txt: bytes, patched_txt: bytes, filename: str
) -> str:
    if compilation_unit.crytic_compile:
        relative_path = compilation_unit.crytic_compile.filename_lookup(filename).relative
        relative_path = os.path.join(".", relative_path)
    else:
        relative_path = filename
    diff = difflib.unified_diff(
        original_txt.decode("utf8").splitlines(False),
        patched_txt.decode("utf8").splitlines(False),
        fromfile=relative_path,
        tofile=relative_path,
        lineterm="",
    )

    return "\n".join(list(diff)) + "\n"



"""
Module detecting usage of `tx.origin` in a conditional node
"""

from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class TxOrigin(AbstractDetector):
    """
    Detect usage of tx.origin in a conditional node
    """

    ARGUMENT = "tx-origin"
    HELP = "Dangerous usage of `tx.origin`"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#dangerous-usage-of-txorigin"
    )

    WIKI_TITLE = "Dangerous usage of `tx.origin`"
    WIKI_DESCRIPTION = "`tx.origin`-based protection can be abused by a malicious contract if a legitimate user interacts with the malicious contract."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract TxOrigin {
    address owner = msg.sender;

    function bug() {
        require(tx.origin == owner);
    }
```
Bob is the owner of `TxOrigin`. Bob calls Eve's contract. Eve's contract calls `TxOrigin` and bypasses the `tx.origin` protection."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Do not use `tx.origin` for authorization."

    @staticmethod
    def _contains_incorrect_tx_origin_use(node: Node) -> bool:
        """
             Check if the node reads tx.origin and doesn't read msg.sender
             Avoid the FP due to (msg.sender == tx.origin)
        Returns:
            (bool)
        """
        solidity_var_read = node.solidity_variables_read
        if solidity_var_read:
            return any(v.name == "tx.origin" for v in solidity_var_read) and all(
                v.name != "msg.sender" for v in solidity_var_read
            )
        return False

    def detect_tx_origin(self, contract: Contract) -> list[tuple[FunctionContract, list[Node]]]:
        ret = []
        for f in contract.functions:
            nodes = f.nodes
            condtional_nodes = [
                n for n in nodes if n.contains_if() or n.contains_require_or_assert()
            ]
            bad_tx_nodes = [
                n for n in condtional_nodes if self._contains_incorrect_tx_origin_use(n)
            ]
            if bad_tx_nodes:
                ret.append((f, bad_tx_nodes))
        return ret

    def _detect(self) -> list[Output]:
        """Detect the functions that use tx.origin in a conditional node"""
        results = []
        for c in self.contracts:
            values = self.detect_tx_origin(c)
            for func, nodes in values:
                for node in nodes:
                    info: DETECTOR_INFO = [func, " uses tx.origin for authorization: ", node, "\n"]
                    res = self.generate_result(info)
                    results.append(res)

        return results



from pathlib import Path
from slither import Slither


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_local_alias(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.18")
    slither = Slither(
        Path(TEST_DATA_DIR, "variable_read_write_storage_pointer/local_alias.sol").as_posix(),
        solc=solc_path,
    )

    contract = slither.get_contract_from_name("Test")[0]

    test = contract.get_function_from_full_name("test()")

    s0 = contract.get_state_variable_from_name("s0")
    s1 = contract.get_state_variable_from_name("s1")

    assert set(test.state_variables_written) == {s0, s1}


def test_parameter_no_library(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.18")
    slither = Slither(
        Path(TEST_DATA_DIR, "variable_read_write_storage_pointer/without_library.sol").as_posix(),
        solc=solc_path,
    )

    contract = slither.get_contract_from_name("MinterRole")[0]

    print(contract.available_functions_as_dict())
    add = contract.get_function_from_full_name("add(MinterRole.Role,address)")
    has = contract.get_function_from_full_name("has(MinterRole.Role,address)")

    minter = contract.get_state_variable_from_name("_minters")

    assert set(add.state_variables_written) == {minter}
    assert set(add.state_variables_read) == {minter}
    assert set(has.state_variables_written) == set()
    assert set(has.state_variables_read) == {minter}



import pytest

from slither.core.cfg.node import Node, NodeType
from slither.core.declarations import (
    Function,
    Contract,
)
from slither.slithir.operations import (
    Return,
)


@pytest.mark.parametrize("legacy", [True, False])
def test_with_explicit_return(slither_from_solidity_source, legacy) -> None:
    source = """
        contract Contract {
            function foo(int x) public returns (int y) {
                if(x > 0) {
                    return x;
                } else {
                    y = x * -1;
                }
            }
        }
        """
    with slither_from_solidity_source(source, legacy=legacy) as slither:
        c: Contract = slither.get_contract_from_name("Contract")[0]
        f: Function = c.functions[0]
        node_if: Node = f.nodes[1]
        node_true = node_if.son_true
        node_false = node_if.son_false
        assert node_true.type == NodeType.RETURN
        assert isinstance(node_true.irs[0], Return)
        assert node_true.irs[0].values[0] == f.get_local_variable_from_name("x")
        assert len(node_true.sons) == 0
        node_end_if = node_false.sons[0]
        assert node_end_if.type == NodeType.ENDIF
        assert node_end_if.sons[0].type == NodeType.RETURN
        node_ret = node_end_if.sons[0]
        assert isinstance(node_ret.irs[0], Return)
        assert node_ret.irs[0].values[0] == f.get_local_variable_from_name("y")


@pytest.mark.parametrize("legacy", [True, False])
def test_return_multiple_with_struct(slither_from_solidity_source, legacy) -> None:
    source = """
        struct St {
            uint256 value;
        }

        contract Contract {
            function foo(St memory x) public returns (St memory y, uint256 z) {
                z = x.value;
                y = St({value: z + 1});
            }
        }
        """
    with slither_from_solidity_source(source, legacy=legacy) as slither:
        c: Contract = slither.get_contract_from_name("Contract")[0]
        f: Function = c.functions[0]
        assert len(f.nodes) == 4
        node = f.nodes[3]
        assert node.type == NodeType.RETURN
        assert isinstance(node.irs[0], Return)
        assert node.irs[0].values[0] == f.get_local_variable_from_name("y")
        assert node.irs[0].values[1] == f.get_local_variable_from_name("z")


def test_nested_ifs_with_loop_legacy(slither_from_solidity_source) -> None:
    source = """
        contract Contract {
            function foo(uint a) public returns (uint x) {
                x = a;
                if(a == 1) {
                    return a;
                } else {
                    for (uint i = 0; i < 10; i++) {
                        if (x > 10) {
                            if (a < 0) {
                                x = 10 * x;
                            } else {
                                throw;
                            }
                        } else {
                            x++;
                        }
                    }
                }
            }
        }
        """
    with slither_from_solidity_source(source, solc_version="0.4.1", legacy=True) as slither:
        c: Contract = slither.get_contract_from_name("Contract")[0]
        f: Function = c.functions[0]
        node_if = f.nodes[2]
        assert node_if.son_true.type == NodeType.RETURN
        node_explicit = node_if.son_true
        assert isinstance(node_explicit.irs[0], Return)
        assert node_explicit.irs[0].values[0] == f.get_local_variable_from_name("a")
        node_end_if = f.nodes[16]
        assert node_end_if.type == NodeType.ENDIF
        assert node_end_if.sons[0].type == NodeType.RETURN
        node_implicit = node_end_if.sons[0]
        assert isinstance(node_implicit.irs[0], Return)
        assert node_implicit.irs[0].values[0] == f.get_local_variable_from_name("x")
        node_throw = f.nodes[11]
        assert node_throw.type == NodeType.THROW
        assert len(node_throw.sons) == 0


def test_nested_ifs_with_loop_compact(slither_from_solidity_source) -> None:
    source = """
        contract Contract {
            function foo(uint a) public returns (uint x) {
                x = a;
                if(a == 1) {
                    return a;
                } else {
                    require(a != 1);
                    for (uint i = 0; i < 10; i++) {
                        if (x > 10) {
                            if (a < 0) {
                                x = 10 * x;
                            } else {
                                revert();
                            }
                        } else {
                            x++;
                        }
                    }
                }
            }
        }
        """
    with slither_from_solidity_source(source, solc_version="0.8.0", legacy=False) as slither:
        c: Contract = slither.get_contract_from_name("Contract")[0]
        f: Function = c.functions[0]
        node_if = f.nodes[2]
        assert node_if.son_true.type == NodeType.RETURN
        node_explicit = node_if.son_true
        assert isinstance(node_explicit.irs[0], Return)
        assert node_explicit.irs[0].values[0] == f.get_local_variable_from_name("a")
        node_end_if = f.nodes[17]
        assert node_end_if.type == NodeType.ENDIF
        assert node_end_if.sons[0].type == NodeType.RETURN
        node_implicit = node_end_if.sons[0]
        assert isinstance(node_implicit.irs[0], Return)
        assert node_implicit.irs[0].values[0] == f.get_local_variable_from_name("x")


@pytest.mark.xfail  # Explicit returns inside assembly are currently not parsed as return nodes
@pytest.mark.parametrize("legacy", [True, False])
def test_assembly_switch_cases(slither_from_solidity_source, legacy):
    source = """
        contract Contract {
            function foo(uint a) public returns (uint x) {
                assembly {
                    switch a
                    case 0 { x := 0 }
                    case 1 { x := 1 }
                    default {
                        return(0x0, 32)
                    }
                }
            }
        }
        """
    with slither_from_solidity_source(source, solc_version="0.8.0", legacy=legacy) as slither:
        c: Contract = slither.get_contract_from_name("Contract")[0]
        f = c.functions[0]
        if legacy:
            node = f.nodes[2]
            assert node.type == NodeType.RETURN
            assert isinstance(node.irs[0], Return)
            assert node.irs[0].values[0] == f.get_local_variable_from_name("x")
        else:
            node_end_if = f.nodes[5]
            assert node_end_if.sons[0].type == NodeType.RETURN
            node_implicit = node_end_if.sons[0]
            assert isinstance(node_implicit.irs[0], Return)
            assert node_implicit.irs[0].values[0] == f.get_local_variable_from_name("x")
            # This part will fail until issue #1927 is fixed
            node_explicit = f.nodes[10]
            assert node_explicit.type == NodeType.RETURN
            assert len(node_explicit.sons) == 0


@pytest.mark.parametrize("legacy", [True, False])
def test_issue_1846_ternary_in_ternary(slither_from_solidity_source, legacy):
    source = """
        contract Contract {
            function foo(uint x) public returns (uint y) {
                y = x > 0 ? x > 1 ? 2 : 3 : 4;
            }
        }
        """
    with slither_from_solidity_source(source, legacy=legacy) as slither:
        c: Contract = slither.get_contract_from_name("Contract")[0]
        f = c.functions[0]
        node_end_if = f.nodes[3]
        assert node_end_if.type == NodeType.ENDIF
        assert len(node_end_if.sons) == 1
        node_ret = node_end_if.sons[0]
        assert node_ret.type == NodeType.RETURN
        assert isinstance(node_ret.irs[0], Return)
        assert node_ret.irs[0].values[0] == f.get_local_variable_from_name("y")



import logging
from functools import lru_cache

from slither.core.expressions.assignment_operation import AssignmentOperation
from slither.core.expressions.binary_operation import BinaryOperation
from slither.core.expressions.call_expression import CallExpression
from slither.core.expressions.conditional_expression import ConditionalExpression
from slither.core.expressions.elementary_type_name_expression import ElementaryTypeNameExpression
from slither.core.expressions.expression import Expression
from slither.core.expressions.identifier import Identifier
from slither.core.expressions.index_access import IndexAccess
from slither.core.expressions.literal import Literal
from slither.core.expressions.member_access import MemberAccess
from slither.core.expressions.new_array import NewArray
from slither.core.expressions.new_contract import NewContract
from slither.core.expressions.new_elementary_type import NewElementaryType
from slither.core.expressions.tuple_expression import TupleExpression
from slither.core.expressions.type_conversion import TypeConversion
from slither.core.expressions.unary_operation import UnaryOperation
from slither.core.expressions.super_call_expression import SuperCallExpression
from slither.core.expressions.super_identifier import SuperIdentifier
from slither.core.expressions.self_identifier import SelfIdentifier
from slither.exceptions import SlitherError

logger = logging.getLogger("ExpressionVisitor")


@lru_cache
def get_visitor_mapping():
    """Returns a visitor mapping from expression type to visiting functions."""
    return {
        AssignmentOperation: "_visit_assignement_operation",
        BinaryOperation: "_visit_binary_operation",
        CallExpression: "_visit_call_expression",
        ConditionalExpression: "_visit_conditional_expression",
        ElementaryTypeNameExpression: "_visit_elementary_type_name_expression",
        Identifier: "_visit_identifier",
        IndexAccess: "_visit_index_access",
        Literal: "_visit_literal",
        MemberAccess: "_visit_member_access",
        NewArray: "_visit_new_array",
        NewContract: "_visit_new_contract",
        NewElementaryType: "_visit_new_elementary_type",
        TupleExpression: "_visit_tuple_expression",
        TypeConversion: "_visit_type_conversion",
        UnaryOperation: "_visit_unary_operation",
        SelfIdentifier: "_visit_identifier",
        SuperIdentifier: "_visit_identifier",
        SuperCallExpression: "_visit_call_expression",
    }


class ExpressionVisitor:
    def __init__(self, expression: Expression) -> None:
        super().__init__()
        # Inherited class must declare their variables prior calling super().__init__
        self._expression = expression
        self._visit_expression(self.expression)

    @property
    def expression(self) -> Expression:
        return self._expression

    # visit an expression
    # call pre_visit, visit_expression_name, post_visit
    def _visit_expression(self, expression: Expression) -> None:
        self._pre_visit(expression)

        if expression is not None:
            visitor_method = get_visitor_mapping().get(expression.__class__)
            if not visitor_method:
                raise SlitherError(f"Expression not handled: {expression}")

            visitor = getattr(self, visitor_method)
            visitor(expression)

        self._post_visit(expression)

    # visit_expression_name

    def _visit_assignement_operation(self, expression: AssignmentOperation) -> None:
        self._visit_expression(expression.expression_left)
        self._visit_expression(expression.expression_right)

    def _visit_binary_operation(self, expression: BinaryOperation) -> None:
        self._visit_expression(expression.expression_left)
        self._visit_expression(expression.expression_right)

    def _visit_call_expression(self, expression: CallExpression) -> None:
        self._visit_expression(expression.called)
        for arg in expression.arguments:
            if arg:
                self._visit_expression(arg)
        if expression.call_value:
            self._visit_expression(expression.call_value)
        if expression.call_gas:
            self._visit_expression(expression.call_gas)
        if expression.call_salt:
            self._visit_expression(expression.call_salt)

    def _visit_conditional_expression(self, expression: ConditionalExpression) -> None:
        self._visit_expression(expression.if_expression)
        self._visit_expression(expression.else_expression)
        self._visit_expression(expression.then_expression)

    def _visit_elementary_type_name_expression(
        self, expression: ElementaryTypeNameExpression
    ) -> None:
        pass

    def _visit_identifier(self, expression: Identifier) -> None:
        pass

    def _visit_index_access(self, expression: IndexAccess) -> None:
        self._visit_expression(expression.expression_left)
        self._visit_expression(expression.expression_right)

    def _visit_literal(self, expression: Literal) -> None:
        pass

    def _visit_member_access(self, expression: MemberAccess) -> None:
        self._visit_expression(expression.expression)

    def _visit_new_array(self, expression: NewArray) -> None:
        pass

    def _visit_new_contract(self, expression: NewContract) -> None:
        pass

    def _visit_new_elementary_type(self, expression: NewElementaryType) -> None:
        pass

    def _visit_tuple_expression(self, expression: TupleExpression) -> None:
        for e in expression.expressions:
            if e:
                self._visit_expression(e)

    def _visit_type_conversion(self, expression: TypeConversion) -> None:
        self._visit_expression(expression.expression)

    def _visit_unary_operation(self, expression: UnaryOperation) -> None:
        self._visit_expression(expression.expression)

    # pre visit

    def _pre_visit(self, expression: Expression) -> None:
        if isinstance(expression, AssignmentOperation):
            self._pre_assignement_operation(expression)

        elif isinstance(expression, BinaryOperation):
            self._pre_binary_operation(expression)

        elif isinstance(expression, CallExpression):
            self._pre_call_expression(expression)

        elif isinstance(expression, ConditionalExpression):
            self._pre_conditional_expression(expression)

        elif isinstance(expression, ElementaryTypeNameExpression):
            self._pre_elementary_type_name_expression(expression)

        elif isinstance(expression, Identifier):
            self._pre_identifier(expression)

        elif isinstance(expression, IndexAccess):
            self._pre_index_access(expression)

        elif isinstance(expression, Literal):
            self._pre_literal(expression)

        elif isinstance(expression, MemberAccess):
            self._pre_member_access(expression)

        elif isinstance(expression, NewArray):
            self._pre_new_array(expression)

        elif isinstance(expression, NewContract):
            self._pre_new_contract(expression)

        elif isinstance(expression, NewElementaryType):
            self._pre_new_elementary_type(expression)

        elif isinstance(expression, TupleExpression):
            self._pre_tuple_expression(expression)

        elif isinstance(expression, TypeConversion):
            self._pre_type_conversion(expression)

        elif isinstance(expression, UnaryOperation):
            self._pre_unary_operation(expression)

        elif expression is None:
            pass

        else:
            raise SlitherError(f"Expression not handled: {expression}")

    # pre_expression_name

    def _pre_assignement_operation(self, expression: AssignmentOperation) -> None:
        pass

    def _pre_binary_operation(self, expression: BinaryOperation) -> None:
        pass

    def _pre_call_expression(self, expression: CallExpression) -> None:
        pass

    def _pre_conditional_expression(self, expression: ConditionalExpression) -> None:
        pass

    def _pre_elementary_type_name_expression(
        self, expression: ElementaryTypeNameExpression
    ) -> None:
        pass

    def _pre_identifier(self, expression: Identifier) -> None:
        pass

    def _pre_index_access(self, expression: IndexAccess) -> None:
        pass

    def _pre_literal(self, expression: Literal) -> None:
        pass

    def _pre_member_access(self, expression: MemberAccess) -> None:
        pass

    def _pre_new_array(self, expression: NewArray) -> None:
        pass

    def _pre_new_contract(self, expression: NewContract) -> None:
        pass

    def _pre_new_elementary_type(self, expression: NewElementaryType) -> None:
        pass

    def _pre_tuple_expression(self, expression: TupleExpression) -> None:
        pass

    def _pre_type_conversion(self, expression: TypeConversion) -> None:
        pass

    def _pre_unary_operation(self, expression: UnaryOperation) -> None:
        pass

    # post visit

    def _post_visit(self, expression: Expression) -> None:
        if isinstance(expression, AssignmentOperation):
            self._post_assignement_operation(expression)

        elif isinstance(expression, BinaryOperation):
            self._post_binary_operation(expression)

        elif isinstance(expression, CallExpression):
            self._post_call_expression(expression)

        elif isinstance(expression, ConditionalExpression):
            self._post_conditional_expression(expression)

        elif isinstance(expression, ElementaryTypeNameExpression):
            self._post_elementary_type_name_expression(expression)

        elif isinstance(expression, Identifier):
            self._post_identifier(expression)

        elif isinstance(expression, IndexAccess):
            self._post_index_access(expression)

        elif isinstance(expression, Literal):
            self._post_literal(expression)

        elif isinstance(expression, MemberAccess):
            self._post_member_access(expression)

        elif isinstance(expression, NewArray):
            self._post_new_array(expression)

        elif isinstance(expression, NewContract):
            self._post_new_contract(expression)

        elif isinstance(expression, NewElementaryType):
            self._post_new_elementary_type(expression)

        elif isinstance(expression, TupleExpression):
            self._post_tuple_expression(expression)

        elif isinstance(expression, TypeConversion):
            self._post_type_conversion(expression)

        elif isinstance(expression, UnaryOperation):
            self._post_unary_operation(expression)

        elif expression is None:
            pass

        else:
            raise SlitherError(f"Expression not handled: {expression}")

    # post_expression_name

    def _post_assignement_operation(self, expression: AssignmentOperation) -> None:
        pass

    def _post_binary_operation(self, expression: BinaryOperation) -> None:
        pass

    def _post_call_expression(self, expression: CallExpression) -> None:
        pass

    def _post_conditional_expression(self, expression: ConditionalExpression) -> None:
        pass

    def _post_elementary_type_name_expression(
        self, expression: ElementaryTypeNameExpression
    ) -> None:
        pass

    def _post_identifier(self, expression: Identifier) -> None:
        pass

    def _post_index_access(self, expression: IndexAccess) -> None:
        pass

    def _post_literal(self, expression: Literal) -> None:
        pass

    def _post_member_access(self, expression: MemberAccess) -> None:
        pass

    def _post_new_array(self, expression: NewArray) -> None:
        pass

    def _post_new_contract(self, expression: NewContract) -> None:
        pass

    def _post_new_elementary_type(self, expression: NewElementaryType) -> None:
        pass

    def _post_tuple_expression(self, expression: TupleExpression) -> None:
        pass

    def _post_type_conversion(self, expression: TypeConversion) -> None:
        pass

    def _post_unary_operation(self, expression: UnaryOperation) -> None:
        pass



from typing import Any

from slither.core.solidity_types.type import Type
from slither.core.variables.function_type_variable import FunctionTypeVariable


class FunctionType(Type):
    def __init__(
        self,
        params: list[FunctionTypeVariable],
        return_values: list[FunctionTypeVariable],
    ) -> None:
        assert all(isinstance(x, FunctionTypeVariable) for x in params)
        assert all(isinstance(x, FunctionTypeVariable) for x in return_values)
        super().__init__()
        self._params: list[FunctionTypeVariable] = params
        self._return_values: list[FunctionTypeVariable] = return_values

    @property
    def params(self) -> list[FunctionTypeVariable]:
        return self._params

    @property
    def return_values(self) -> list[FunctionTypeVariable]:
        return self._return_values

    @property
    def return_type(self) -> list[Type]:
        return [x.type for x in self.return_values]

    @property
    def storage_size(self) -> tuple[int, bool]:
        return 24, False

    @property
    def is_dynamic(self) -> bool:
        return False

    def __str__(self) -> str:
        # Use x.type
        # x.name may be empty
        params = ",".join([str(x.type) for x in self._params])
        return_values = ",".join([str(x.type) for x in self._return_values])
        if return_values:
            return f"function({params}) returns({return_values})"
        return f"function({params})"

    @property
    def parameters_signature(self) -> str:
        """
        Return the parameters signature(without the return statetement)
        """
        # Use x.type
        # x.name may be empty
        params = ",".join([str(x.type) for x in self._params])
        return f"({params})"

    @property
    def signature(self) -> str:
        """
        Return the signature(with the return statetement if it exists)
        """
        # Use x.type
        # x.name may be empty
        params = ",".join([str(x.type) for x in self._params])
        return_values = ",".join([str(x.type) for x in self._return_values])
        if return_values:
            return f"({params}) returns({return_values})"
        return f"({params})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FunctionType):
            return False
        return self.params == other.params and self.return_values == other.return_values

    def __hash__(self):
        return hash(str(self))



from slither.slithir.operations import Binary, BinaryType
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator
from slither.core.expressions.unary_operation import UnaryOperation
from slither.core.expressions.call_expression import CallExpression
from slither.core.expressions.member_access import MemberAccess

arithmetic_operators = [
    BinaryType.ADDITION,
    BinaryType.DIVISION,
    BinaryType.MULTIPLICATION,
    BinaryType.SUBTRACTION,
    BinaryType.MODULO,
]


class AOR(AbstractMutator):
    NAME = "AOR"
    HELP = "Arithmetic operator replacement"

    def _mutate(self) -> dict:
        result: dict = {}
        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                try:
                    ir_expression = node.expression
                except AttributeError:
                    continue

                # Special cases handling .push and .pop on dynamic arrays.
                # The IR for these operations has a binary operation due to internal conversion
                # (see convert_to_push and convert_to_pop in slithir/convert.py)
                # however it's not present in the source code and should not be mutated.

                if (
                    isinstance(ir_expression, CallExpression)
                    and isinstance(ir_expression.called, MemberAccess)
                    and ir_expression.called.member_name == "pop"
                ):
                    continue

                # For a .push instruction we skip the last 6 IR operations
                # because they are fixed based on the internal conversion to the IR
                # while we need to look at the preceding instructions because
                # they might contain Binary IR to be mutated.
                # For example for a.push(3+x) it's correct to mutate 3+x.
                irs = (
                    node.irs[:-6]
                    if isinstance(ir_expression, CallExpression)
                    and isinstance(ir_expression.called, MemberAccess)
                    and ir_expression.called.member_name == "push"
                    else node.irs
                )

                for ir in irs:
                    if isinstance(ir, Binary) and ir.type in arithmetic_operators:
                        if isinstance(ir_expression, UnaryOperation):
                            continue
                        alternative_ops = arithmetic_operators[:]
                        alternative_ops.remove(ir.type)
                        for op in alternative_ops:
                            # Get the string
                            start = node.source_mapping.start
                            stop = start + node.source_mapping.length
                            old_str = node.source_mapping.content
                            line_no = node.source_mapping.lines
                            halves = old_str.split(ir.type.value)
                            if len(halves) != 2:
                                continue  # skip if assembly
                            new_str = f"{halves[0]}{op.value}{halves[1]}"
                            create_patch_with_line(
                                result,
                                self.in_file,
                                start,
                                stop,
                                old_str,
                                new_str,
                                line_no[0],
                            )
        return result



import sys
from slither.slither import Slither
from slither.slithir.convert import convert_expression


if len(sys.argv) != 2:
    print("python function_called.py functions_called.sol")
    sys.exit(-1)

# Init slither
slither = Slither(sys.argv[1])

# Get the contract
contracts = slither.get_contract_from_name("Test")
assert len(contracts) == 1
contract = contracts[0]
# Get the variable
test = contract.get_function_from_signature("one()")
assert test
nodes = test.nodes

for node in nodes:
    if node.expression:
        print(f"Expression:\n\t{node.expression}")
        irs = convert_expression(node.expression, node)
        print("IR expressions:")
        for ir in irs:
            print(f"\t{ir}")
        print()



"""
Module detecting uninitialized local variables

Recursively explore the CFG to only report uninitialized local variables that are
read before being written
"""

from slither.core.cfg.node import Node, NodeType
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.utils.output import Output


class UninitializedLocalVars(AbstractDetector):
    ARGUMENT = "uninitialized-local"
    HELP = "Uninitialized local variables"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#uninitialized-local-variables"

    WIKI_TITLE = "Uninitialized local variables"
    WIKI_DESCRIPTION = "Uninitialized local variables."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Uninitialized is Owner{
    function withdraw() payable public onlyOwner{
        address to;
        to.transfer(this.balance)
    }
}
```
Bob calls `transfer`. As a result, all Ether is sent to the address `0x0` and is lost."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Initialize all the variables. If a variable is meant to be initialized to zero, explicitly set it to zero to improve code readability."

    key = "UNINITIALIZEDLOCAL"

    def _detect_uninitialized(
        self, function: FunctionContract, node: Node, visited: list[Node]
    ) -> None:
        if node in visited:
            return

        visited = visited + [node]

        fathers_context = []

        for father in node.fathers:
            if self.key in father.context:
                fathers_context += father.context[self.key]

        # Exclude path that dont bring further information
        if node in self.visited_all_paths:
            if all(f_c in self.visited_all_paths[node] for f_c in fathers_context):
                return
        else:
            self.visited_all_paths[node] = []

        self.visited_all_paths[node] = list(set(self.visited_all_paths[node] + fathers_context))

        # Remove a local variable declared in a for loop header
        if (
            node.type == NodeType.VARIABLE
            and len(node.sons) == 1  # Should always be true for a node that has a STARTLOOP son
            and node.sons[0].type == NodeType.STARTLOOP
        ):
            if node.variable_declaration in fathers_context:
                fathers_context.remove(node.variable_declaration)

        if self.key in node.context:
            fathers_context += node.context[self.key]

        variables_read = node.variables_read
        for uninitialized_local_variable in fathers_context:
            if uninitialized_local_variable in variables_read:
                self.results.append((function, uninitialized_local_variable))

        # Only save the local variables that are not yet written
        uninitialized_local_variables = list(set(fathers_context) - set(node.variables_written))
        node.context[self.key] = uninitialized_local_variables

        for son in node.sons:
            self._detect_uninitialized(function, son, visited)

    def _detect(self) -> list[Output]:
        """Detect uninitialized local variables

        Recursively visit the calls
        Returns:
            dict: [contract name] = set(local variable uninitialized)
        """
        results = []

        self.results = []
        self.visited_all_paths = {}

        for contract in self.compilation_unit.contracts:
            for function in contract.functions:
                if (
                    function.is_implemented
                    and function.contract_declarer == contract
                    and function.entry_point
                ):
                    if function.contains_assembly:
                        continue
                    # dont consider storage variable, as they are detected by another detector
                    uninitialized_local_variables = [
                        v for v in function.local_variables if not v.is_storage and v.uninitialized
                    ]
                    function.entry_point.context[self.key] = uninitialized_local_variables
                    self._detect_uninitialized(function, function.entry_point, [])
        all_results = list(set(self.results))
        for function, uninitialized_local_variable in all_results:
            info = [
                uninitialized_local_variable,
                " is a local variable never initialized\n",
            ]
            json = self.generate_result(info)
            results.append(json)

        return results



from slither.slithir.operations.internal_call import InternalCall
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class GelatoUnprotectedRandomness(AbstractDetector):
    """
    Unprotected Gelato VRF requests
    """

    ARGUMENT = "gelato-unprotected-randomness"
    HELP = "Call to _requestRandomness within an unprotected function"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#gelato-unprotected-randomness"

    WIKI_TITLE = "Gelato unprotected randomness"
    WIKI_DESCRIPTION = "Detect calls to `_requestRandomness` within an unprotected function."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C is GelatoVRFConsumerBase {
    function _fulfillRandomness(
        uint256 randomness,
        uint256,
        bytes memory extraData
    ) internal override {
        // Do something with the random number
    }

    function bad() public {
        _requestRandomness(abi.encode(msg.sender));
    }
}
```
The function `bad` is uprotected and requests randomness."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        "Function that request randomness should be allowed only to authorized users."
    )

    def _detect(self) -> list[Output]:
        results = []

        for contract in self.compilation_unit.contracts_derived:
            if "GelatoVRFConsumerBase" in [c.name for c in contract.inheritance]:
                for function in contract.functions_entry_points:
                    if not function.is_protected() and (
                        nodes_request := [
                            ir.node
                            for ir in function.all_internal_calls()
                            if isinstance(ir, InternalCall)
                            and ir.function_name == "_requestRandomness"
                        ]
                    ):
                        # Sort so output is deterministic
                        nodes_request.sort(key=lambda x: (x.node_id, x.function.full_name))

                        for node in nodes_request:
                            info: DETECTOR_INFO = [
                                function,
                                " is unprotected and request randomness from Gelato VRF\n\t- ",
                                node,
                                "\n",
                            ]
                            res = self.generate_result(info)
                            results.append(res)

        return results



from pathlib import Path
from typing import TYPE_CHECKING

from slither.core.source_mapping.source_mapping import SourceMapping

if TYPE_CHECKING:
    from slither.core.scope.scope import FileScope


class Import(SourceMapping):
    def __init__(self, filename: Path, scope: "FileScope") -> None:
        super().__init__()
        self._filename: Path = filename
        self.scope: FileScope = scope
        self._alias: str | None = None
        # Map local name -> original name
        self.renaming: dict[str, str] = {}

        self._pattern = "import"

    @property
    def filename(self) -> str:
        """
        Return the absolute filename

        :return:
        :rtype:
        """
        return self._filename.as_posix()

    @property
    def filename_path(self) -> Path:
        """
        Return the absolute filename

        :return:
        :rtype:
        """

        return self._filename

    @property
    def alias(self) -> str | None:
        return self._alias

    @alias.setter
    def alias(self, a: str):
        self._alias = a

    def __str__(self):
        return self.filename



from functools import total_ordering

from slither.core.solidity_types.elementary_type import ElementaryType, Int, Uint
from slither.slithir.variables.variable import SlithIRVariable
from slither.utils.arithmetic import convert_subdenomination
from slither.utils.integer_conversion import convert_string_to_int


@total_ordering
class Constant(SlithIRVariable):
    def __init__(
        self,
        val: str,
        constant_type: ElementaryType | None = None,
        subdenomination: str | None = None,
    ) -> None:
        super().__init__()
        assert isinstance(val, str)

        self._original_value = val
        self._subdenomination = subdenomination

        if subdenomination:
            val = str(convert_subdenomination(val, subdenomination))

        if constant_type:
            assert isinstance(constant_type, ElementaryType)
            self._type = constant_type
            if constant_type.type in Int + Uint + ["address"]:
                self._val: bool | int | str = convert_string_to_int(val)
            elif constant_type.type == "bool":
                self._val = (val == "true") | (val == "True")
            else:
                self._val = val
        else:
            if val.isdigit():
                self._type = ElementaryType("uint256")
                self._val = convert_string_to_int(val)
            else:
                self._type = ElementaryType("string")
                self._val = val

        self._name = str(self._val)

    @property
    def value(self) -> bool | int | str:
        """
        Return the value.
        If the expression was an hexadecimal delcared as hex'...'
        return a str
        Returns:
            (str | int | bool)
        """
        return self._val

    @property
    def original_value(self) -> str:
        """
        Return the string representation of the value
        :return: str
        """
        return self._original_value

    def __str__(self) -> str:
        return str(self.value)

    def __eq__(self, other: object) -> bool:
        return self.value == other

    def __ne__(self, other: object) -> bool:
        return self.value != other

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, (Constant, str)):
            raise NotImplementedError
        return self.value < other  # type: ignore

    def __repr__(self) -> str:
        return f"{self.value!s}"

    def __hash__(self) -> int:
        return self._val.__hash__()



from pathlib import Path
from collections import namedtuple
from slither import Slither
from slither.slithir.operations import Operation, NewContract

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def check_num_local_vars_read(function, slithir_op: Operation, num_reads_expected: int):
    for node in function.nodes:
        for operation in node.irs:
            if isinstance(operation, slithir_op):
                assert len(operation.read) == num_reads_expected
                assert len(node.local_variables_read) == num_reads_expected


def check_num_states_vars_read(function, slithir_op: Operation, num_reads_expected: int):
    for node in function.nodes:
        for operation in node.irs:
            if isinstance(operation, slithir_op):
                assert len(operation.read) == num_reads_expected
                assert len(node.state_variables_read) == num_reads_expected


OperationTest = namedtuple("OperationTest", "contract_name slithir_op")

OPERATION_TEST = [OperationTest("NewContract", NewContract)]


def test_operation_reads(solc_binary_path) -> None:
    """
    Every slithir operation has its own contract and reads all local and state variables in readAllLocalVariables and readAllStateVariables, respectively.
    """
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(Path(TEST_DATA_DIR, "operation_reads.sol").as_posix(), solc=solc_path)

    for op_test in OPERATION_TEST:
        print(op_test)
        available = slither.get_contract_from_name(op_test.contract_name)
        assert len(available) == 1
        target = available[0]

        num_state_variables = len(target.state_variables_ordered)
        state_function = target.get_function_from_signature("readAllStateVariables()")
        check_num_states_vars_read(state_function, op_test.slithir_op, num_state_variables)

        local_function = target.get_function_from_signature("readAllLocalVariables()")
        num_local_vars = len(local_function.local_variables)
        check_num_local_vars_read(local_function, op_test.slithir_op, num_local_vars)



import re
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator


function_header_replacements = [
    "pure ==> view",
    "view ==> pure",
    "(\\s)(external|public|internal) ==> \\1private",
    "(\\s)(external|public) ==> \\1internal",
]


class FHR(AbstractMutator):
    NAME = "FHR"
    HELP = "Function Header Replacement"

    def _mutate(self) -> dict:
        result: dict = {}

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            start = function.source_mapping.start
            stop = start + function.source_mapping.content.find("{")
            old_str = function.source_mapping.content
            line_no = function.source_mapping.lines
            if line_no[0] not in self.dont_mutate_line:
                for value in function_header_replacements:
                    left_value = value.split(" ==> ", maxsplit=1)[0]
                    right_value = value.split(" ==> ")[1]
                    if re.search(re.compile(left_value), old_str) is not None:
                        new_str = re.sub(re.compile(left_value), right_value, old_str)
                        create_patch_with_line(
                            result,
                            self.in_file,
                            start,
                            stop,
                            old_str,
                            new_str,
                            line_no[0],
                        )
        return result
